import time
from typing import Optional

import psutil

from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.message import Resource, WorkerHeartbeat, WorkerHeartbeatEcho
from scaler.protocol.python.status import ProcessorStatus
from scaler.utility.mixins import Looper
from scaler.worker.agent.mixins import HeartbeatManager, ProcessorManager, TaskManager, TimeoutManager
from scaler.worker.agent.processor_holder import ProcessorHolder


class VanillaHeartbeatManager(Looper, HeartbeatManager):
    def __init__(self):
        self._agent_process = psutil.Process()

        self._connector_external: Optional[AsyncConnector] = None
        self._worker_task_manager: Optional[TaskManager] = None
        self._timeout_manager: Optional[TimeoutManager] = None
        self._processor_manager: Optional[ProcessorManager] = None

        self._start_timestamp_ns = 0
        self._latency_us = 0

    def register(
        self,
        connector_external: AsyncConnector,
        worker_task_manager: TaskManager,
        timeout_manager: TimeoutManager,
        processor_manager: ProcessorManager,
    ):
        self._connector_external = connector_external
        self._worker_task_manager = worker_task_manager
        self._timeout_manager = timeout_manager
        self._processor_manager = processor_manager

    async def on_heartbeat_echo(self, heartbeat: WorkerHeartbeatEcho):
        if self._start_timestamp_ns == 0:
            # not handling echo if we didn't send out heartbeat
            return

        self._latency_us = int(((time.time_ns() - self._start_timestamp_ns) / 2) // 1_000)
        self._start_timestamp_ns = 0
        self._timeout_manager.update_last_seen_time()

    async def routine(self):
        processors = self._processor_manager.processors()

        if len(processors) == 0:
            return

        if self._start_timestamp_ns != 0:
            # already sent heartbeat, expecting heartbeat echo, so not sending
            return

        for processor_holder in processors:
            status = processor_holder.process().status()
            if status in {psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD}:
                await self._processor_manager.on_failing_processor(processor_holder.processor_id(), status)

        processors = self._processor_manager.processors()  # refreshes for removed dead and zombie processors
        num_suspended_processors = self._processor_manager.num_suspended_processors()

        await self._connector_external.send(
            WorkerHeartbeat.new_msg(
                Resource.new_msg(int(self._agent_process.cpu_percent() * 10), self._agent_process.memory_info().rss),
                psutil.virtual_memory().available,
                self._worker_task_manager.get_queued_size() - num_suspended_processors,
                self._latency_us,
                self._processor_manager.can_accept_task(),
                [self.__get_processor_status_from_holder(processor) for processor in processors],
            )
        )
        self._start_timestamp_ns = time.time_ns()

    @staticmethod
    def __get_processor_status_from_holder(processor: ProcessorHolder) -> ProcessorStatus:
        process = processor.process()

        try:
            resource = Resource.new_msg(int(process.cpu_percent() * 10), process.memory_info().rss)
        except psutil.ZombieProcess:
            # Assumes dead processes do not use any resources
            resource = Resource.new_msg(0, 0)

        return ProcessorStatus.new_msg(
            processor.pid(), processor.initialized(), processor.task() is not None, processor.suspended(), resource
        )



from typing import Optional, Set

from scaler.client.agent.future_manager import ClientFutureManager
from scaler.client.agent.mixins import ObjectManager, TaskManager
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.message import GraphTask, GraphTaskCancel, Task, TaskCancel, TaskResult


class ClientTaskManager(TaskManager):
    def __init__(self):
        self._task_ids: Set[bytes] = set()
        self._graph_task_ids: Set[bytes] = set()

        self._connector_external: Optional[AsyncConnector] = None
        self._object_manager: Optional[ObjectManager] = None
        self._future_manager: Optional[ClientFutureManager] = None

    def register(
        self, connector_external: AsyncConnector, object_manager: ObjectManager, future_manager: ClientFutureManager
    ):
        self._connector_external = connector_external
        self._object_manager = object_manager
        self._future_manager = future_manager

    async def on_new_task(self, task: Task):
        self._task_ids.add(task.task_id)
        await self._connector_external.send(task)

    async def on_cancel_task(self, task_cancel: TaskCancel):
        # We might receive a cancel task event on a previously finished task if:
        # -	The scheduler sends a TaskResult message to the client agent
        # - The client sends a TaskCancel message to the client agent, as it's not yet aware the task finished.
        # - The client agent processes the TaskResult message and removes the task from self._task_ids
        # - The client agent processes the TaskCancel message (that was already queued before processing the
        #   TaskResult), and fails on self._task_ids.remove() as the task_id no longer exists.

        if task_cancel.task_id not in self._task_ids:
            return

        self._task_ids.remove(task_cancel.task_id)
        self._future_manager.on_cancel_task(task_cancel)

        await self._connector_external.send(task_cancel)

    async def on_new_graph_task(self, task: GraphTask):
        self._graph_task_ids.add(task.task_id)
        self._task_ids.update(set(task.targets))
        await self._connector_external.send(task)

    async def on_cancel_graph_task(self, task_cancel: GraphTaskCancel):
        if task_cancel.task_id not in self._graph_task_ids:
            return

        self._graph_task_ids.remove(task_cancel.task_id)
        await self._connector_external.send(task_cancel)

    async def on_task_result(self, result: TaskResult):
        # All task result objects must be propagated to the object manager, even if we do not track the task anymore
        # (e.g. if it got cancelled). If we don't, we might lose track of these result objects and not properly clear
        # them.
        self._object_manager.on_task_result(result)

        if result.task_id not in self._task_ids:
            return

        self._task_ids.remove(result.task_id)

        self._future_manager.on_task_result(result)



import dataclasses
import enum
import uuid
from asyncio import Queue
from typing import Dict, List, Optional, Set

from scaler.io.async_binder import AsyncBinder
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.common import ObjectContent, TaskStatus
from scaler.protocol.python.message import GraphTask, GraphTaskCancel, StateGraphTask, Task, TaskCancel, TaskResult
from scaler.scheduler.mixins import ClientManager, GraphTaskManager, ObjectManager, TaskManager
from scaler.utility.graph.topological_sorter import TopologicalSorter
from scaler.utility.many_to_many_dict import ManyToManyDict
from scaler.utility.mixins import Looper, Reporter


class _NodeTaskState(enum.Enum):
    Inactive = enum.auto()
    Running = enum.auto()
    Canceled = enum.auto()
    Failed = enum.auto()
    Success = enum.auto()


class _GraphState(enum.Enum):
    Running = enum.auto()
    Canceling = enum.auto()


@dataclasses.dataclass
class _TaskInfo:
    state: _NodeTaskState
    task: Task
    result_object_ids: List[bytes] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class _Graph:
    target_task_ids: List[bytes]
    sorter: TopologicalSorter
    tasks: Dict[bytes, _TaskInfo]
    depended_task_id_to_task_id: ManyToManyDict[bytes, bytes]
    client: bytes
    status: _GraphState = dataclasses.field(default=_GraphState.Running)
    running_task_ids: Set[bytes] = dataclasses.field(default_factory=set)


class VanillaGraphTaskManager(GraphTaskManager, Looper, Reporter):
    """
    A = func()
    B = func2(A)
    C = func3(A)
    D = func4(B, C)

    graph
    A = Task(func)
    B = Task(func2, A)
    C = Task(func3, A)
    D = Task(func4, B, C)

    dependencies
    {"A": {B, C}
     "B": {D},
     "C": {D},
     "D": {},
    }
    """

    def __init__(self):
        self._binder: Optional[AsyncBinder] = None
        self._binder_monitor: Optional[AsyncConnector] = None
        self._client_manager: Optional[ClientManager] = None
        self._task_manager: Optional[TaskManager] = None
        self._object_manager: Optional[ObjectManager] = None

        self._unassigned: Queue = Queue()

        self._graph_task_id_to_graph: Dict[bytes, _Graph] = dict()
        self._task_id_to_graph_task_id: Dict[bytes, bytes] = dict()

    def register(
        self,
        binder: AsyncBinder,
        binder_monitor: AsyncConnector,
        client_manager: ClientManager,
        task_manager: TaskManager,
        object_manager: ObjectManager,
    ):
        self._binder = binder
        self._binder_monitor = binder_monitor
        self._client_manager = client_manager
        self._task_manager = task_manager
        self._object_manager = object_manager

    async def on_graph_task(self, client: bytes, graph_task: GraphTask):
        await self._unassigned.put((client, graph_task))

    async def on_graph_task_cancel(self, client: bytes, graph_task_cancel: GraphTaskCancel):
        if graph_task_cancel.task_id not in self._graph_task_id_to_graph:
            await self._binder.send(client, TaskResult.new_msg(graph_task_cancel.task_id, TaskStatus.NotFound))
            return

        graph_task_id = self._task_id_to_graph_task_id[graph_task_cancel.task_id]
        graph_info = self._graph_task_id_to_graph[graph_task_id]
        if graph_info.status == _GraphState.Canceling:
            return

        await self.__cancel_one_graph(graph_task_id, TaskResult.new_msg(graph_task_cancel.task_id, TaskStatus.Canceled))

    async def on_graph_sub_task_done(self, result: TaskResult):
        graph_task_id = self._task_id_to_graph_task_id[result.task_id]
        graph_info = self._graph_task_id_to_graph[graph_task_id]
        if graph_info.status == _GraphState.Canceling:
            return

        await self.__mark_node_done(result)

        if result.status == TaskStatus.Success:
            await self.__check_one_graph(graph_task_id)
            return

        assert result.status != TaskStatus.Success
        await self.__cancel_one_graph(graph_task_id, result)

    def is_graph_sub_task(self, task_id: bytes):
        return task_id in self._task_id_to_graph_task_id

    async def routine(self):
        client, graph_task = await self._unassigned.get()
        await self.__add_new_graph(client, graph_task)

    def get_status(self) -> Dict:
        return {"graph_manager": {"unassigned": self._unassigned.qsize()}}

    async def __add_new_graph(self, client: bytes, graph_task: GraphTask):
        graph = {}

        self._client_manager.on_task_begin(client, graph_task.task_id)

        tasks = dict()
        depended_task_id_to_task_id: ManyToManyDict[bytes, bytes] = ManyToManyDict()
        for task in graph_task.graph:
            self._task_id_to_graph_task_id[task.task_id] = graph_task.task_id
            tasks[task.task_id] = _TaskInfo(_NodeTaskState.Inactive, task)

            required_task_ids = {arg.data for arg in task.function_args if arg.type == Task.Argument.ArgumentType.Task}
            for required_task_id in required_task_ids:
                depended_task_id_to_task_id.add(required_task_id, task.task_id)

            graph[task.task_id] = required_task_ids

            await self._binder_monitor.send(
                StateGraphTask.new_msg(
                    graph_task.task_id,
                    task.task_id,
                    (
                        StateGraphTask.NodeTaskType.Target
                        if task.task_id in graph_task.targets
                        else StateGraphTask.NodeTaskType.Normal
                    ),
                    required_task_ids,
                )
            )

        sorter = TopologicalSorter(graph)
        sorter.prepare()

        self._graph_task_id_to_graph[graph_task.task_id] = _Graph(
            graph_task.targets, sorter, tasks, depended_task_id_to_task_id, client
        )
        await self.__check_one_graph(graph_task.task_id)

    async def __check_one_graph(self, graph_task_id: bytes):
        graph_info = self._graph_task_id_to_graph[graph_task_id]
        if not graph_info.sorter.is_active():
            await self.__finish_one_graph(graph_task_id, TaskResult.new_msg(graph_task_id, TaskStatus.Success))
            return

        ready_task_ids = graph_info.sorter.get_ready()
        if not ready_task_ids:
            return

        for task_id in ready_task_ids:
            task_info = graph_info.tasks[task_id]
            task_info.state = _NodeTaskState.Running
            graph_info.running_task_ids.add(task_id)

            task = Task.new_msg(
                task_id=task_info.task.task_id,
                source=task_info.task.source,
                metadata=task_info.task.metadata,
                func_object_id=task_info.task.func_object_id,
                function_args=[self.__get_argument(graph_task_id, arg) for arg in task_info.task.function_args],
            )

            await self._task_manager.on_task_new(graph_info.client, task)

    async def __mark_node_done(self, result: TaskResult):
        graph_task_id = self._task_id_to_graph_task_id.pop(result.task_id)

        graph_info = self._graph_task_id_to_graph[graph_task_id]

        task_info = graph_info.tasks[result.task_id]

        task_info.result_object_ids = result.results

        if result.status == TaskStatus.Success:
            task_info.state = _NodeTaskState.Success
        elif result.status == TaskStatus.Canceled:
            task_info.state = _NodeTaskState.Canceled
        elif result.status == TaskStatus.Failed:
            task_info.state = _NodeTaskState.Failed
        elif result.status == TaskStatus.NotFound:
            task_info.state = _NodeTaskState.Canceled

        else:
            raise ValueError(f"received unexpected task result {result}")

        self.__clean_intermediate_result(graph_task_id, result.task_id)
        graph_info.sorter.done(result.task_id)

        if result.task_id in graph_info.running_task_ids:
            graph_info.running_task_ids.remove(result.task_id)

        if result.task_id in graph_info.target_task_ids:
            await self._binder.send(graph_info.client, result)

    async def __cancel_one_graph(self, graph_task_id: bytes, result: TaskResult):
        graph_info = self._graph_task_id_to_graph[graph_task_id]
        graph_info.status = _GraphState.Canceling

        if not self.__is_graph_finished(graph_task_id):
            await self.__clean_all_running_nodes(graph_task_id, result)
            await self.__clean_all_inactive_nodes(graph_task_id, result)

        await self.__finish_one_graph(
            graph_task_id, TaskResult.new_msg(result.task_id, result.status, result.metadata, result.results)
        )

    async def __clean_all_running_nodes(self, graph_task_id: bytes, result: TaskResult):
        graph_info = self._graph_task_id_to_graph[graph_task_id]

        running_task_ids = graph_info.running_task_ids.copy()

        # cancel all running tasks
        for task_id in running_task_ids:
            new_result_object_ids = []
            for result_object_id in result.results:
                new_result_object_id = uuid.uuid4().bytes
                self._object_manager.on_add_object(
                    graph_info.client,
                    new_result_object_id,
                    ObjectContent.ObjectContentType.Object,
                    self._object_manager.get_object_name(result_object_id),
                    self._object_manager.get_object_content(result_object_id),
                )
                new_result_object_ids.append(new_result_object_id)

            await self._task_manager.on_task_cancel(graph_info.client, TaskCancel.new_msg(task_id))
            await self.__mark_node_done(
                TaskResult.new_msg(task_id, result.status, result.metadata, new_result_object_ids)
            )

    async def __clean_all_inactive_nodes(self, graph_task_id: bytes, result: TaskResult):
        graph_info = self._graph_task_id_to_graph[graph_task_id]
        while graph_info.sorter.is_active():
            ready_task_ids = graph_info.sorter.get_ready()
            for task_id in ready_task_ids:
                new_result_object_ids = []
                for result_object_id in result.results:
                    new_result_object_id = uuid.uuid4().bytes
                    self._object_manager.on_add_object(
                        graph_info.client,
                        new_result_object_id,
                        ObjectContent.ObjectContentType.Object,
                        self._object_manager.get_object_name(result_object_id),
                        self._object_manager.get_object_content(result_object_id),
                    )
                    new_result_object_ids.append(new_result_object_id)

                await self.__mark_node_done(
                    TaskResult.new_msg(task_id, result.status, result.metadata, new_result_object_ids)
                )

    async def __finish_one_graph(self, graph_task_id: bytes, result: TaskResult):
        self._client_manager.on_task_finish(graph_task_id)
        info = self._graph_task_id_to_graph.pop(graph_task_id)
        await self._binder.send(info.client, TaskResult.new_msg(graph_task_id, result.status, results=result.results))

    def __is_graph_finished(self, graph_task_id: bytes):
        graph_info = self._graph_task_id_to_graph[graph_task_id]
        return not graph_info.sorter.is_active() and not graph_info.running_task_ids

    def __get_target_results_ids(self, graph_task_id: bytes) -> List[bytes]:
        graph_info = self._graph_task_id_to_graph[graph_task_id]
        return [
            result_object_id
            for task_id in graph_info.target_task_ids
            for result_object_id in graph_info.tasks[task_id].result_object_ids
        ]

    def __get_argument(self, graph_task_id: bytes, argument: Task.Argument) -> Task.Argument:
        if argument.type == Task.Argument.ArgumentType.ObjectID:
            return argument

        assert argument.type == Task.Argument.ArgumentType.Task
        argument_task_id = argument.data

        graph_info = self._graph_task_id_to_graph[graph_task_id]
        task_info = graph_info.tasks[argument_task_id]

        assert len(task_info.result_object_ids) == 1

        return Task.Argument(Task.Argument.ArgumentType.ObjectID, task_info.result_object_ids[0])

    def __clean_intermediate_result(self, graph_task_id: bytes, task_id: bytes):
        graph_info = self._graph_task_id_to_graph[graph_task_id]
        task_info = graph_info.tasks[task_id]

        for argument in filter(lambda arg: arg.type == Task.Argument.ArgumentType.Task, task_info.task.function_args):
            argument_task_id = argument.data
            graph_info.depended_task_id_to_task_id.remove(argument_task_id, task_id)
            if graph_info.depended_task_id_to_task_id.has_left_key(argument_task_id):
                continue

            # delete intermediate results as they are not needed anymore
            self._object_manager.on_del_objects(
                graph_info.client, set(graph_info.tasks[argument_task_id].result_object_ids)
            )



import abc
from typing import List, Optional, Set

from scaler.protocol.python.common import ObjectContent
from scaler.protocol.python.message import (
    ClientDisconnect,
    ClientHeartbeat,
    DisconnectRequest,
    GraphTask,
    GraphTaskCancel,
    ObjectInstruction,
    ObjectRequest,
    Task,
    TaskCancel,
    TaskResult,
    WorkerHeartbeat,
)
from scaler.utility.mixins import Reporter


class ObjectManager(Reporter):
    @abc.abstractmethod
    async def on_object_instruction(self, source: bytes, request: ObjectInstruction):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_object_request(self, source: bytes, request: ObjectRequest):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_add_object(
        self,
        object_user: bytes,
        object_id: bytes,
        object_type: ObjectContent.ObjectContentType,
        object_name: bytes,
        object_bytes: List[bytes],
    ):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_del_objects(self, task_id: bytes, object_ids: Set[bytes]):
        raise NotImplementedError()

    @abc.abstractmethod
    def clean_client(self, client: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    def has_object(self, object_id: bytes) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_object_name(self, object_id: bytes) -> bytes:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_object_content(self, object_id: bytes) -> List[bytes]:
        raise NotImplementedError()


class ClientManager(Reporter):
    @abc.abstractmethod
    def get_client_task_ids(self, client: bytes) -> Set[bytes]:
        raise NotImplementedError()

    @abc.abstractmethod
    def has_client_id(self, client_id: bytes) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_client_id(self, task_id: bytes) -> Optional[bytes]:
        raise NotImplementedError()

    @abc.abstractmethod
    def on_task_begin(self, client: bytes, task_id: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_task_finish(self, task_id: bytes) -> bytes:
        raise NotImplementedError()

    async def on_heartbeat(self, client: bytes, info: ClientHeartbeat):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_client_disconnect(self, client: bytes, request: ClientDisconnect):
        raise NotImplementedError()


class GraphTaskManager(Reporter):
    @abc.abstractmethod
    async def on_graph_task(self, client: bytes, graph_task: GraphTask):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_graph_task_cancel(self, client: bytes, graph_task_cancel: GraphTaskCancel):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_graph_sub_task_done(self, result: TaskResult) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def is_graph_sub_task(self, task_id: bytes):
        raise NotImplementedError()


class TaskManager(Reporter):
    @abc.abstractmethod
    async def on_task_new(self, client: bytes, task: Task):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_cancel(self, client: bytes, task_cancel: TaskCancel):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_done(self, result: TaskResult):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_reroute(self, task_id: bytes):
        raise NotImplementedError()


class WorkerManager(Reporter):
    @abc.abstractmethod
    async def assign_task_to_worker(self, task: Task) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_cancel(self, task_cancel: TaskCancel):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_result(self, task_result: TaskResult):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_heartbeat(self, worker: bytes, info: WorkerHeartbeat):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_client_shutdown(self, client: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_disconnect(self, source: bytes, request: DisconnectRequest):
        raise NotImplementedError()

    @abc.abstractmethod
    def has_available_worker(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_worker_by_task_id(self, task_id: bytes) -> bytes:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_worker_ids(self) -> Set[bytes]:
        raise NotImplementedError()



import asyncio
import enum
import logging
from typing import Awaitable, Callable


class EventLoopType(enum.Enum):
    builtin = enum.auto()
    uvloop = enum.auto()

    @staticmethod
    def allowed_types():
        return {m.name for m in EventLoopType}


def register_event_loop(event_loop_type: str):
    if event_loop_type not in EventLoopType.allowed_types():
        raise TypeError(f"allowed event loop types are: {EventLoopType.allowed_types()}")

    event_loop_type_enum = EventLoopType[event_loop_type]
    if event_loop_type_enum == EventLoopType.uvloop:
        try:
            import uvloop  # noqa
        except ImportError:
            raise ImportError("please use pip install uvloop if try to use uvloop as event loop")

        uvloop.install()

    assert event_loop_type in EventLoopType.allowed_types()

    logging.info(f"use event loop: {event_loop_type}")


def create_async_loop_routine(routine: Callable[[], Awaitable], seconds: int):
    async def loop():
        logging.info(f"{routine.__self__.__class__.__name__}: started")  # type: ignore[attr-defined]
        try:
            while True:
                await routine()
                await asyncio.sleep(seconds)
        except asyncio.CancelledError:
            pass
        except KeyboardInterrupt:
            pass

        logging.info(f"{routine.__self__.__class__.__name__}: exited")  # type: ignore[attr-defined]

    return loop()



import dataclasses
from typing import Any, Dict, Hashable, Optional


@dataclasses.dataclass
class _Node:
    value: Any
    prev: Optional["_Node"] = None
    next: Optional["_Node"] = None


class _DoubleLinkedList:
    def __init__(self):
        self._head: Optional[_Node] = None
        self._tail: Optional[_Node] = None
        self._size = 0

    def __len__(self):
        return self._size

    def add_to_head(self, node: _Node):
        if self._head is None:
            self._head = node
            self._tail = node
        else:
            node.next = self._head
            self._head.prev = node
            self._head = node

        self._size += 1

    def remove_tail(self):
        if self._tail is None:
            raise IndexError(f"{self.__class__.__name__} queue empty")

        node = self._tail
        if self._tail.prev is None:
            self._head = None
            self._tail = None
        else:
            self._tail = self._tail.prev
            self._tail.next = None

        self._size -= 1
        return node

    def remove(self, node: _Node):
        prev_node = node.prev
        next_node = node.next
        if prev_node and next_node:
            prev_node.next = next_node
            next_node.prev = prev_node

        elif not prev_node and not next_node:
            assert self._head is node
            assert self._tail is node
            self._head = None
            self._tail = None

        elif prev_node and not next_node:
            assert self._tail is node
            prev_node.next = None
            self._tail = prev_node

        elif not prev_node and next_node:
            assert self._head is node
            next_node.prev = None
            self._head = next_node

        self._size -= 1
        del node


class IndexedQueue:
    """A queue that provides O(1) operations for adding and removing any item."""

    def __init__(self):
        self._double_linked_list = _DoubleLinkedList()
        self._hash_map: Dict[int, _Node] = {}

    def __contains__(self, item: Hashable):
        key = hash(item)
        return key in self._hash_map

    def __len__(self):
        return self._double_linked_list.__len__()

    def __iter__(self):
        node = self._double_linked_list._tail
        while node is not None:
            yield node.value
            node = node.prev

    def put(self, item: Hashable):
        key = hash(item)
        if key in self._hash_map:
            raise KeyError(f"{self.__class__.__name__} already have item: {item}")

        node = _Node(item)
        self._double_linked_list.add_to_head(node)
        self._hash_map[key] = node

    def get(self):
        node = self._double_linked_list.remove_tail()
        del self._hash_map[hash(node.value)]
        return node.value

    def remove(self, item: Hashable):
        key = hash(item)
        if key not in self._hash_map:
            raise ValueError(f"{self.__class__.__name__} doesn't have item: {item}")

        node = self._hash_map.pop(key)
        self._double_linked_list.remove(node)



import logging
import os
import uuid
from typing import Awaitable, Callable, List, Literal, Optional

import zmq.asyncio

from scaler.io.utility import deserialize, serialize
from scaler.protocol.python.mixins import Message
from scaler.utility.zmq_config import ZMQConfig


class AsyncConnector:
    def __init__(
        self,
        context: zmq.asyncio.Context,
        name: str,
        socket_type: int,
        address: ZMQConfig,
        bind_or_connect: Literal["bind", "connect"],
        callback: Optional[Callable[[Message], Awaitable[None]]],
        identity: Optional[bytes],
    ):
        self._address = address

        self._context = context
        self._socket = self._context.socket(socket_type)

        if identity is None:
            identity = f"{os.getpid()}|{name}|{uuid.uuid4().bytes.hex()}".encode()
        self._identity = identity

        # set socket option
        self._socket.setsockopt(zmq.IDENTITY, self._identity)
        self._socket.setsockopt(zmq.SNDHWM, 0)
        self._socket.setsockopt(zmq.RCVHWM, 0)

        if bind_or_connect == "bind":
            self._socket.bind(self._address.to_address())
        elif bind_or_connect == "connect":
            self._socket.connect(self._address.to_address())
        else:
            raise TypeError("bind_or_connect has to be 'bind' or 'connect'")

        self._callback: Optional[Callable[[Message], Awaitable[None]]] = callback

    def __del__(self):
        self.destroy()

    def destroy(self):
        if self._socket.closed:
            return

        self._socket.close(linger=1)

    @property
    def identity(self) -> bytes:
        return self._identity

    @property
    def socket(self) -> zmq.asyncio.Socket:
        return self._socket

    @property
    def address(self) -> str:
        return self._address.to_address()

    async def routine(self):
        if self._callback is None:
            return

        message: Optional[Message] = await self.receive()
        if message is None:
            return

        await self._callback(message)

    async def receive(self) -> Optional[Message]:
        if self._context.closed:
            return None

        if self._socket.closed:
            return None

        payload = await self._socket.recv(copy=False)
        result: Optional[Message] = deserialize(payload.bytes)
        if result is None:
            logging.error(f"received unknown message: {payload.bytes!r}")
            return None

        return result

    async def send(self, message: Message):
        await self._socket.send(serialize(message), copy=False)

    def __is_valid_message(self, frames: List[bytes]) -> bool:
        if len(frames) > 1:
            logging.error(f"{self.__get_prefix()} received unexpected frames {frames}")
            return False

        return True

    def __get_prefix(self):
        return f"{self.__class__.__name__}[{self._identity.decode()}]:"



import dataclasses
import enum
import os
from typing import List, Optional, Set, Tuple, Type

import bidict

from scaler.protocol.capnp._python import _message  # noqa
from scaler.protocol.python.common import ObjectContent, TaskStatus
from scaler.protocol.python.mixins import Message
from scaler.protocol.python.status import (
    BinderStatus,
    ClientManagerStatus,
    ObjectManagerStatus,
    ProcessorStatus,
    Resource,
    TaskManagerStatus,
    WorkerManagerStatus,
)


class Task(Message):
    @dataclasses.dataclass
    class Argument:
        type: "ArgumentType"
        data: bytes

        def __repr__(self):
            return f"Argument(type={self.type}, data={self.data.hex()})"

        class ArgumentType(enum.Enum):
            Task = _message.Task.Argument.ArgumentType.task
            ObjectID = _message.Task.Argument.ArgumentType.objectID

    def __init__(self, msg):
        super().__init__(msg)

    def __repr__(self):
        return (
            f"Task(task_id={self.task_id.hex()}, source={self.source.hex()}, metadata={self.metadata.hex()},"
            f"func_object_id={self.func_object_id.hex()}, function_args={self.function_args})"
        )

    @property
    def task_id(self) -> bytes:
        return self._msg.taskId

    @property
    def source(self) -> bytes:
        return self._msg.source

    @property
    def metadata(self) -> bytes:
        return self._msg.metadata

    @property
    def func_object_id(self) -> bytes:
        return self._msg.funcObjectId

    @property
    def function_args(self) -> List[Argument]:
        return [
            Task.Argument(type=Task.Argument.ArgumentType(arg.type.raw), data=arg.data)
            for arg in self._msg.functionArgs
        ]

    @staticmethod
    def new_msg(
        task_id: bytes, source: bytes, metadata: bytes, func_object_id: bytes, function_args: List[Argument]
    ) -> "Task":
        return Task(
            _message.Task(
                taskId=task_id,
                source=source,
                metadata=metadata,
                funcObjectId=func_object_id,
                functionArgs=[_message.Task.Argument(type=arg.type.value, data=arg.data) for arg in function_args],
            )
        )


class TaskCancel(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @dataclasses.dataclass
    class TaskCancelFlags:
        force: bool
        retrieve_task_object: bool

    @property
    def task_id(self) -> bytes:
        return self._msg.taskId

    @property
    def flags(self) -> TaskCancelFlags:
        return TaskCancel.TaskCancelFlags(
            force=self._msg.flags.force, retrieve_task_object=self._msg.flags.retrieveTaskObject
        )

    @staticmethod
    def new_msg(task_id: bytes, flags: Optional[TaskCancelFlags] = None) -> "TaskCancel":
        if flags is None:
            flags = TaskCancel.TaskCancelFlags(force=False, retrieve_task_object=False)

        return TaskCancel(
            _message.TaskCancel(
                taskId=task_id,
                flags=_message.TaskCancel.TaskCancelFlags(
                    force=flags.force, retrieveTaskObject=flags.retrieve_task_object
                ),
            )
        )


class TaskResult(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def task_id(self) -> bytes:
        return self._msg.taskId

    @property
    def status(self) -> TaskStatus:
        return TaskStatus(self._msg.status.raw)

    @property
    def metadata(self) -> bytes:
        return self._msg.metadata

    @property
    def results(self) -> List[bytes]:
        return self._msg.results

    @staticmethod
    def new_msg(
        task_id: bytes, status: TaskStatus, metadata: Optional[bytes] = None, results: Optional[List[bytes]] = None
    ) -> "TaskResult":
        if metadata is None:
            metadata = bytes()

        if results is None:
            results = list()

        return TaskResult(_message.TaskResult(taskId=task_id, status=status.value, metadata=metadata, results=results))


class GraphTask(Message):
    def __init__(self, msg):
        super().__init__(msg)

    def __repr__(self):
        return (
            f"GraphTask({os.linesep}"
            f"    task_id={self.task_id.hex()},{os.linesep}"
            f"    targets=[{os.linesep}"
            f"        {[target.hex() + ',' + os.linesep for target in self.targets]}"
            f"    ]\n"
            f"    graph={self.graph}\n"
            f")"
        )

    @property
    def task_id(self) -> bytes:
        return self._msg.taskId

    @property
    def source(self) -> bytes:
        return self._msg.source

    @property
    def targets(self) -> List[bytes]:
        return self._msg.targets

    @property
    def graph(self) -> List[Task]:
        return [Task(task) for task in self._msg.graph]

    @staticmethod
    def new_msg(task_id: bytes, source: bytes, targets: List[bytes], graph: List[Task]) -> "GraphTask":
        return GraphTask(
            _message.GraphTask(
                taskId=task_id, source=source, targets=targets, graph=[task.get_message() for task in graph]
            )
        )


class GraphTaskCancel(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def task_id(self) -> bytes:
        return self._msg.taskId

    @staticmethod
    def new_msg(task_id: bytes) -> "GraphTaskCancel":
        return GraphTaskCancel(_message.GraphTaskCancel(taskId=task_id))

    def get_message(self):
        return self._msg


class ClientHeartbeat(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def resource(self) -> Resource:
        return Resource(self._msg.resource)

    @property
    def latency_us(self) -> int:
        return self._msg.latencyUS

    @staticmethod
    def new_msg(resource: Resource, latency_us: int) -> "ClientHeartbeat":
        return ClientHeartbeat(_message.ClientHeartbeat(resource=resource.get_message(), latencyUS=latency_us))


class ClientHeartbeatEcho(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @staticmethod
    def new_msg() -> "ClientHeartbeatEcho":
        return ClientHeartbeatEcho(_message.ClientHeartbeatEcho())


class WorkerHeartbeat(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def agent(self) -> Resource:
        return Resource(self._msg.agent)

    @property
    def rss_free(self) -> int:
        return self._msg.rssFree

    @property
    def queued_tasks(self) -> int:
        return self._msg.queuedTasks

    @property
    def latency_us(self) -> int:
        return self._msg.latencyUS

    @property
    def task_lock(self) -> bool:
        return self._msg.taskLock

    @property
    def processors(self) -> List[ProcessorStatus]:
        return [ProcessorStatus(p) for p in self._msg.processors]

    @staticmethod
    def new_msg(
        agent: Resource,
        rss_free: int,
        queued_tasks: int,
        latency_us: int,
        task_lock: bool,
        processors: List[ProcessorStatus],
    ) -> "WorkerHeartbeat":
        return WorkerHeartbeat(
            _message.WorkerHeartbeat(
                agent=agent.get_message(),
                rssFree=rss_free,
                queuedTasks=queued_tasks,
                latencyUS=latency_us,
                taskLock=task_lock,
                processors=[p.get_message() for p in processors],
            )
        )


class WorkerHeartbeatEcho(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @staticmethod
    def new_msg() -> "WorkerHeartbeatEcho":
        return WorkerHeartbeatEcho(_message.WorkerHeartbeatEcho())


class ObjectInstruction(Message):
    class ObjectInstructionType(enum.Enum):
        Create = _message.ObjectInstruction.ObjectInstructionType.create
        Delete = _message.ObjectInstruction.ObjectInstructionType.delete
        Clear = _message.ObjectInstruction.ObjectInstructionType.clear

    def __init__(self, msg):
        super().__init__(msg)

    @property
    def instruction_type(self) -> ObjectInstructionType:
        return ObjectInstruction.ObjectInstructionType(self._msg.instructionType.raw)

    @property
    def object_user(self) -> bytes:
        return self._msg.objectUser

    @property
    def object_content(self) -> ObjectContent:
        return ObjectContent(self._msg.objectContent)

    @staticmethod
    def new_msg(
        instruction_type: ObjectInstructionType, object_user: bytes, object_content: ObjectContent
    ) -> "ObjectInstruction":
        return ObjectInstruction(
            _message.ObjectInstruction(
                instructionType=instruction_type.value,
                objectUser=object_user,
                objectContent=object_content.get_message(),
            )
        )


class ObjectRequest(Message):
    class ObjectRequestType(enum.Enum):
        Get = _message.ObjectRequest.ObjectRequestType.get

    def __init__(self, msg):
        super().__init__(msg)

    def __repr__(self):
        return (
            f"ObjectRequest(type={self.request_type}, "
            f"object_ids={tuple(object_id.hex() for object_id in self.object_ids)})"
        )

    @property
    def request_type(self) -> ObjectRequestType:
        return ObjectRequest.ObjectRequestType(self._msg.requestType.raw)

    @property
    def object_ids(self) -> Tuple[bytes]:
        return tuple(self._msg.objectIds)

    @staticmethod
    def new_msg(request_type: ObjectRequestType, object_ids: Tuple[bytes, ...]) -> "ObjectRequest":
        return ObjectRequest(_message.ObjectRequest(requestType=request_type.value, objectIds=list(object_ids)))


class ObjectResponse(Message):
    class ObjectResponseType(enum.Enum):
        Content = _message.ObjectResponse.ObjectResponseType.content
        ObjectNotExist = _message.ObjectResponse.ObjectResponseType.objectNotExist

    def __init__(self, msg):
        super().__init__(msg)

    @property
    def response_type(self) -> ObjectResponseType:
        return ObjectResponse.ObjectResponseType(self._msg.responseType.raw)

    @property
    def object_content(self) -> ObjectContent:
        return ObjectContent(self._msg.objectContent)

    @staticmethod
    def new_msg(response_type: ObjectResponseType, object_content: ObjectContent) -> "ObjectResponse":
        return ObjectResponse(
            _message.ObjectResponse(responseType=response_type.value, objectContent=object_content.get_message())
        )


class DisconnectRequest(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def worker(self) -> bytes:
        return self._msg.worker

    @staticmethod
    def new_msg(worker: bytes) -> "DisconnectRequest":
        return DisconnectRequest(_message.DisconnectRequest(worker=worker))


@dataclasses.dataclass
class DisconnectResponse(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def worker(self) -> bytes:
        return self._msg.worker

    @staticmethod
    def new_msg(worker: bytes) -> "DisconnectResponse":
        return DisconnectResponse(_message.DisconnectResponse(worker=worker))


class ClientDisconnect(Message):
    class DisconnectType(enum.Enum):
        Disconnect = _message.ClientDisconnect.DisconnectType.disconnect
        Shutdown = _message.ClientDisconnect.DisconnectType.shutdown

    def __init__(self, msg):
        super().__init__(msg)

    @property
    def disconnect_type(self) -> DisconnectType:
        return ClientDisconnect.DisconnectType(self._msg.disconnectType.raw)

    @staticmethod
    def new_msg(disconnect_type: DisconnectType) -> "ClientDisconnect":
        return ClientDisconnect(_message.ClientDisconnect(disconnectType=disconnect_type.value))


class ClientShutdownResponse(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def accepted(self) -> bool:
        return self._msg.accepted

    @staticmethod
    def new_msg(accepted: bool) -> "ClientShutdownResponse":
        return ClientShutdownResponse(_message.ClientShutdownResponse(accepted=accepted))


class StateClient(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @staticmethod
    def new_msg() -> "StateClient":
        return StateClient(_message.StateClient())


class StateObject(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @staticmethod
    def new_msg() -> "StateObject":
        return StateObject(_message.StateObject())


class StateBalanceAdvice(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def worker_id(self) -> bytes:
        return self._msg.workerId

    @property
    def task_ids(self) -> List[bytes]:
        return self._msg.taskIds

    @staticmethod
    def new_msg(worker_id: bytes, task_ids: List[bytes]) -> "StateBalanceAdvice":
        return StateBalanceAdvice(_message.StateBalanceAdvice(workerId=worker_id, taskIds=task_ids))


class StateScheduler(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def binder(self) -> BinderStatus:
        return BinderStatus(self._msg.binder)

    @property
    def scheduler(self) -> Resource:
        return Resource(self._msg.scheduler)

    @property
    def rss_free(self) -> int:
        return self._msg.rssFree

    @property
    def client_manager(self) -> ClientManagerStatus:
        return ClientManagerStatus(self._msg.clientManager)

    @property
    def object_manager(self) -> ObjectManagerStatus:
        return ObjectManagerStatus(self._msg.objectManager)

    @property
    def task_manager(self) -> TaskManagerStatus:
        return TaskManagerStatus(self._msg.taskManager)

    @property
    def worker_manager(self) -> WorkerManagerStatus:
        return WorkerManagerStatus(self._msg.workerManager)

    @staticmethod
    def new_msg(
        binder: BinderStatus,
        scheduler: Resource,
        rss_free: int,
        client_manager: ClientManagerStatus,
        object_manager: ObjectManagerStatus,
        task_manager: TaskManagerStatus,
        worker_manager: WorkerManagerStatus,
    ) -> "StateScheduler":
        return StateScheduler(
            _message.StateScheduler(
                binder=binder.get_message(),
                scheduler=scheduler.get_message(),
                rssFree=rss_free,
                clientManager=client_manager.get_message(),
                objectManager=object_manager.get_message(),
                taskManager=task_manager.get_message(),
                workerManager=worker_manager.get_message(),
            )
        )


class StateWorker(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def worker_id(self) -> bytes:
        return self._msg.workerId

    @property
    def message(self) -> bytes:
        return self._msg.message

    @staticmethod
    def new_msg(worker_id: bytes, message: bytes) -> "StateWorker":
        return StateWorker(_message.StateWorker(workerId=worker_id, message=message))


class StateTask(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def task_id(self) -> bytes:
        return self._msg.taskId

    @property
    def function_name(self) -> bytes:
        return self._msg.functionName

    @property
    def status(self) -> TaskStatus:
        return TaskStatus(self._msg.status.raw)

    @property
    def worker(self) -> bytes:
        return self._msg.worker

    @property
    def metadata(self) -> bytes:
        return self._msg.metadata

    @staticmethod
    def new_msg(
        task_id: bytes, function_name: bytes, status: TaskStatus, worker: bytes, metadata: bytes = b""
    ) -> "StateTask":
        return StateTask(
            _message.StateTask(
                taskId=task_id, functionName=function_name, status=status.value, worker=worker, metadata=metadata
            )
        )


class StateGraphTask(Message):
    class NodeTaskType(enum.Enum):
        Normal = _message.StateGraphTask.NodeTaskType.normal
        Target = _message.StateGraphTask.NodeTaskType.target

    def __init__(self, msg):
        super().__init__(msg)

    @property
    def graph_task_id(self) -> bytes:
        return self._msg.graphTaskId

    @property
    def task_id(self) -> bytes:
        return self._msg.taskId

    @property
    def node_task_type(self) -> NodeTaskType:
        return StateGraphTask.NodeTaskType(self._msg.nodeTaskType.raw)

    @property
    def parent_task_ids(self) -> Set[bytes]:
        return set(self._msg.parentTaskIds)

    @staticmethod
    def new_msg(
        graph_task_id: bytes, task_id: bytes, node_task_type: NodeTaskType, parent_task_ids: Set[bytes]
    ) -> "StateGraphTask":
        return StateGraphTask(
            _message.StateGraphTask(
                graphTaskId=graph_task_id,
                taskId=task_id,
                nodeTaskType=node_task_type.value,
                parentTaskIds=list(parent_task_ids),
            )
        )


class ProcessorInitialized(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @staticmethod
    def new_msg() -> "ProcessorInitialized":
        return ProcessorInitialized(_message.ProcessorInitialized())


PROTOCOL: bidict.bidict[str, Type[Message]] = bidict.bidict(
    {
        "task": Task,
        "taskCancel": TaskCancel,
        "taskResult": TaskResult,
        "graphTask": GraphTask,
        "graphTaskCancel": GraphTaskCancel,
        "objectInstruction": ObjectInstruction,
        "objectRequest": ObjectRequest,
        "objectResponse": ObjectResponse,
        "clientHeartbeat": ClientHeartbeat,
        "clientHeartbeatEcho": ClientHeartbeatEcho,
        "workerHeartbeat": WorkerHeartbeat,
        "workerHeartbeatEcho": WorkerHeartbeatEcho,
        "disconnectRequest": DisconnectRequest,
        "disconnectResponse": DisconnectResponse,
        "stateClient": StateClient,
        "stateObject": StateObject,
        "stateBalanceAdvice": StateBalanceAdvice,
        "stateScheduler": StateScheduler,
        "stateWorker": StateWorker,
        "stateTask": StateTask,
        "stateGraphTask": StateGraphTask,
        "clientDisconnect": ClientDisconnect,
        "clientShutdownResponse": ClientShutdownResponse,
        "processorInitialized": ProcessorInitialized,
    }
)



import random

from scaler.client.client import Client
from scaler.cluster.combo import SchedulerClusterCombo
from scaler.utility.logging.scoped_logger import ScopedLogger
from scaler.utility.logging.utility import setup_logger


def sleep_print(sec: int):
    return sec * 1


def main():
    setup_logger()

    address = "tcp://127.0.0.1:2345"

    cluster = SchedulerClusterCombo(address=address, n_workers=10, per_worker_queue_size=2, event_loop="builtin")
    client = Client(address=address)

    tasks = [random.randint(0, 101) for _ in range(10000)]

    with ScopedLogger(f"Scaler submit {len(tasks)} tasks"):
        futures = [client.submit(sleep_print, a) for a in tasks]

    with ScopedLogger(f"Scaler gather {len(futures)} results"):
        results = [future.result() for future in futures]

    assert results == tasks

    client.disconnect()
    cluster.shutdown()


if __name__ == "__main__":
    main()



import abc
from typing import Dict, List, Optional, Set

from scaler.protocol.python.message import (
    ObjectInstruction,
    ObjectRequest,
    ObjectResponse,
    Task,
    TaskCancel,
    TaskResult,
    WorkerHeartbeatEcho,
)
from scaler.utility.metadata.profile_result import ProfileResult
from scaler.worker.agent.processor_holder import ProcessorHolder


class HeartbeatManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def on_heartbeat_echo(self, heartbeat: WorkerHeartbeatEcho):
        raise NotImplementedError()


class TimeoutManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def update_last_seen_time(self):
        raise NotImplementedError()


class TaskManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def on_task_new(self, task: Task):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_result(self, result: TaskResult):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_cancel_task(self, task_cancel: TaskCancel):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_queued_size(self):
        raise NotImplementedError()


class ProcessorManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def on_object_instruction(self, instruction: ObjectInstruction):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_object_response(self, request: ObjectResponse):
        raise NotImplementedError()

    @abc.abstractmethod
    def can_accept_task(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    async def wait_until_can_accept_task(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task(self, task: Task) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_cancel_task(self, task_id: bytes) -> Optional[Task]:
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_failing_processor(self, processor_id: bytes, process_status: str):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_suspend_task(self, task_id: bytes) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def on_resume_task(self, task_id: bytes) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def destroy(self, reason: str):
        raise NotImplementedError()

    @abc.abstractmethod
    def initialized(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def current_task(self) -> Optional[Task]:
        raise NotImplementedError()

    @abc.abstractmethod
    def current_task_id(self) -> bytes:
        raise NotImplementedError()

    @abc.abstractmethod
    def processors(self) -> List[ProcessorHolder]:
        raise NotImplementedError()

    @abc.abstractmethod
    def num_suspended_processors(self) -> int:
        raise NotImplementedError()


class ProfilingManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def on_process_start(self, pid: int):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_process_end(self, pid: int):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_task_start(self, pid: int, task_id: bytes):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_task_end(self, pid: int, task_id: bytes) -> ProfileResult:
        raise NotImplementedError()


class ObjectTracker(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def on_object_request(self, processor_id: bytes, object_request: ObjectRequest) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def on_object_response(self, object_response: ObjectResponse) -> Set[bytes]:
        raise NotImplementedError()

    @abc.abstractmethod
    def on_object_instruction(self, object_instruction: ObjectInstruction) -> Dict[bytes, ObjectInstruction]:
        raise NotImplementedError()

    @abc.abstractmethod
    def on_processor_end(self, processor_id: bytes) -> None:
        raise NotImplementedError()



import argparse
import logging
import os
import signal
import socket

from scaler.io.config import (
    DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
    DEFAULT_IO_THREADS,
    DEFAULT_NUMBER_OF_WORKER,
    DEFAULT_WORKER_DEATH_TIMEOUT,
)
from scaler.utility.event_loop import EventLoopType, register_event_loop
from scaler.utility.logging.utility import setup_logger
from scaler.utility.zmq_config import ZMQConfig
from scaler.worker.symphony.worker import SymphonyWorker


def get_args():
    parser = argparse.ArgumentParser(
        "standalone symphony cluster", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--base-concurrency", "-n", type=int, default=DEFAULT_NUMBER_OF_WORKER, help="base task concurrency"
    )
    parser.add_argument(
        "--worker-name", "-w", type=str, default=None, help="worker name, if not specified, it will be hostname"
    )
    parser.add_argument(
        "--heartbeat-interval",
        "-hi",
        type=int,
        default=DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
        help="number of seconds to send heartbeat interval",
    )
    parser.add_argument(
        "--death-timeout-seconds", "-ds", type=int, default=DEFAULT_WORKER_DEATH_TIMEOUT, help="death timeout seconds"
    )
    parser.add_argument(
        "--event-loop", "-el", default="builtin", choices=EventLoopType.allowed_types(), help="select event loop type"
    )
    parser.add_argument(
        "--io-threads", "-it", default=DEFAULT_IO_THREADS, help="specify number of io threads per worker"
    )
    parser.add_argument(
        "--logging-paths",
        "-lp",
        nargs="*",
        type=str,
        default=("/dev/stdout",),
        help='specify where cluster log should logged to, it can be multiple paths, "/dev/stdout" is default for '
        "standard output, each worker will have its own log file with process id appended to the path",
    )
    parser.add_argument(
        "--logging-level",
        "-ll",
        type=str,
        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        default="INFO",
        help="specify the logging level",
    )
    parser.add_argument(
        "--logging-config-file",
        type=str,
        default=None,
        help="use standard python the .conf file the specify python logging file configuration format, this will "
        "bypass --logging-paths and --logging-level at the same time, and this will not work on per worker logging",
    )
    parser.add_argument("address", type=ZMQConfig.from_string, help="scheduler address to connect to")
    parser.add_argument("service_name", type=str, help="symphony service name")
    return parser.parse_args()


def main():
    args = get_args()
    register_event_loop(args.event_loop)

    if args.worker_name is None:
        args.worker_name = f"{socket.gethostname().split('.')[0]}"

    setup_logger(args.logging_paths, args.logging_config_file, args.logging_level)

    worker = SymphonyWorker(
        address=args.address,
        name=args.worker_name,
        service_name=args.service_name,
        base_concurrency=args.base_concurrency,
        heartbeat_interval_seconds=args.heartbeat_interval,
        death_timeout_seconds=args.death_timeout_seconds,
        event_loop=args.event_loop,
        io_threads=args.io_threads,
    )

    def destroy(*args):
        assert args is not None
        logging.info(f"{SymphonyWorker.__class__.__name__}: shutting down Symphony worker[{worker.pid}]")
        os.kill(worker.pid, signal.SIGINT)

    signal.signal(signal.SIGINT, destroy)
    signal.signal(signal.SIGTERM, destroy)

    worker.start()
    logging.info("Symphony worker started")

    worker.join()
    logging.info("Symphony worker stopped")



import dataclasses
import threading
from functools import partial

from nicegui import ui

from scaler.io.sync_subscriber import SyncSubscriber
from scaler.protocol.python.message import StateScheduler, StateTask
from scaler.protocol.python.mixins import Message
from scaler.ui.live_display import SchedulerSection, WorkersSection
from scaler.ui.memory_window import MemoryChart
from scaler.ui.setting_page import Settings
from scaler.ui.task_graph import TaskStream
from scaler.ui.task_log import TaskLogTable
from scaler.ui.worker_processors import WorkerProcessors
from scaler.utility.formatter import format_bytes, format_percentage
from scaler.utility.zmq_config import ZMQConfig


@dataclasses.dataclass
class Sections:
    scheduler_section: SchedulerSection
    workers_section: WorkersSection
    task_stream_section: TaskStream
    memory_usage_section: MemoryChart
    tasklog_section: TaskLogTable
    worker_processors: WorkerProcessors
    settings_section: Settings


def start_webui(address: str, host: str, port: int):
    tables = Sections(
        scheduler_section=SchedulerSection(),
        workers_section=WorkersSection(),
        task_stream_section=TaskStream(),
        memory_usage_section=MemoryChart(),
        tasklog_section=TaskLogTable(),
        worker_processors=WorkerProcessors(),
        settings_section=Settings(),
    )

    with ui.tabs().classes("w-full h-full") as tabs:
        live_tab = ui.tab("Live")
        tasklog_tab = ui.tab("Task Log")
        stream_tab = ui.tab("Worker Task Stream")
        worker_processors_tab = ui.tab("Worker Processors")
        settings_tab = ui.tab("Settings")

    with ui.tab_panels(tabs, value=live_tab).classes("w-full"):
        with ui.tab_panel(live_tab):
            tables.scheduler_section.draw_section()
            tables.workers_section.draw_section()

        with ui.tab_panel(tasklog_tab):
            tables.tasklog_section.draw_section()
            ui.timer(0.5, tables.tasklog_section.draw_section.refresh, active=True)
            pass

        with ui.tab_panel(stream_tab):
            tables.task_stream_section.setup_task_stream(tables.settings_section)
            ui.timer(0.1, tables.task_stream_section.update_plot, active=True)

            tables.memory_usage_section.setup_memory_chart(tables.settings_section)
            ui.timer(0.1, tables.memory_usage_section.update_plot, active=True)

        with ui.tab_panel(worker_processors_tab) as tab_handler:
            tables.worker_processors.draw_section(tab_handler)

        with ui.tab_panel(settings_tab):
            tables.settings_section.draw_section()

    subscriber = SyncSubscriber(
        address=ZMQConfig.from_string(address),
        callback=partial(__show_status, tables=tables),
        topic=b"",
        timeout_seconds=-1,
    )
    subscriber.start()

    ui_thread = threading.Thread(target=partial(ui.run, host=host, port=port, reload=False), daemon=False)
    ui_thread.start()


def __show_status(status: Message, tables: Sections):
    if isinstance(status, StateScheduler):
        __update_scheduler_state(status, tables)
        return

    if isinstance(status, StateTask):
        tables.task_stream_section.handle_task_state(status)
        tables.memory_usage_section.handle_task_state(status)
        tables.tasklog_section.handle_task_state(status)
        return


def __update_scheduler_state(data: StateScheduler, tables: Sections):
    tables.scheduler_section.cpu = format_percentage(data.scheduler.cpu)
    tables.scheduler_section.rss = format_bytes(data.scheduler.rss)
    tables.scheduler_section.rss_free = format_bytes(data.rss_free)

    previous_workers = set(tables.workers_section.workers.keys())
    current_workers = set(worker_data.worker_id.decode() for worker_data in data.worker_manager.workers)

    for worker_data in data.worker_manager.workers:
        worker_name = worker_data.worker_id.decode()
        tables.workers_section.workers[worker_name].populate(worker_data)

    for died_worker in previous_workers - current_workers:
        tables.workers_section.workers.pop(died_worker)

    if previous_workers != current_workers:
        tables.workers_section.draw_section.refresh()

    tables.task_stream_section.update_data(tables.workers_section)

    tables.worker_processors.update_data(data.worker_manager.workers)



import dataclasses
from collections import defaultdict
from typing import Dict, List, Optional

from nicegui import ui
from nicegui.element import Element

from scaler.protocol.python.status import WorkerStatus
from scaler.ui.utility import format_worker_name
from scaler.utility.formatter import format_microseconds, format_seconds


@dataclasses.dataclass
class SchedulerSection:
    cpu: str = dataclasses.field(default="")
    rss: str = dataclasses.field(default="")
    rss_free: str = dataclasses.field(default="")

    handler: Optional[Element] = dataclasses.field(default=None)

    def draw_section(self):
        with ui.card().classes("w-full"), ui.row() as handler:
            self.handler = handler
            ui.label("Scheduler")
            ui.label()
            ui.label("CPU:")
            ui.label().bind_text_from(self, "cpu")
            ui.label()
            ui.label("RSS:")
            ui.label().bind_text_from(self, "rss")
            ui.label()
            ui.label("RSS Free:")
            ui.label().bind_text_from(self, "rss_free")

    def delete_section(self):
        self.handler.clear()
        self.handler.delete()


@dataclasses.dataclass
class WorkerRow:
    worker: str = dataclasses.field(default="")
    agt_cpu: float = dataclasses.field(default=0)
    agt_rss: int = dataclasses.field(default=0)
    cpu: float = dataclasses.field(default=0)
    rss: int = dataclasses.field(default=0)
    rss_free: int = dataclasses.field(default=0)
    free: int = dataclasses.field(default=0)
    sent: int = dataclasses.field(default=0)
    queued: int = dataclasses.field(default=0)
    suspended: int = dataclasses.field(default=0)
    lag: str = dataclasses.field(default="")
    itl: str = dataclasses.field(default="")
    last_seen: str = dataclasses.field(default="")

    handlers: List[Element] = dataclasses.field(default_factory=list)

    def populate(self, data: WorkerStatus):
        self.worker = data.worker_id.decode()
        self.agt_cpu = data.agent.cpu / 10
        self.agt_rss = int(data.agent.rss / 1e6)
        self.cpu = sum(p.resource.cpu for p in data.processor_statuses) / 10
        self.rss = int(sum(p.resource.rss for p in data.processor_statuses) / 1e6)
        self.rss_free = int(data.rss_free / 1e6)
        self.free = data.free
        self.sent = data.sent
        self.queued = data.queued
        self.suspended = data.suspended
        self.lag = format_microseconds(data.lag_us)
        self.itl = data.itl
        self.last_seen = format_seconds(data.last_s)

    def draw_row(self):
        total_rss = self.rss + self.rss_free

        ui.label(format_worker_name(self.worker))
        ui.knob(track_color="grey-2", show_value=True, min=0, max=100).bind_value_from(self, "agt_cpu")
        ui.knob(track_color="grey-2", show_value=True, min=0, max=total_rss).bind_value_from(self, "agt_rss")
        ui.knob(track_color="grey-2", show_value=True, min=0, max=100).bind_value_from(self, "cpu")
        ui.knob(track_color="grey-2", show_value=True, min=0, max=total_rss).bind_value_from(self, "rss")
        ui.label().bind_text_from(self, "free")
        ui.label().bind_text_from(self, "sent")
        ui.label().bind_text_from(self, "queued")
        ui.label().bind_text_from(self, "suspended")
        ui.label().bind_text_from(self, "lag")
        ui.label().bind_text_from(self, "ITL")
        ui.label().bind_text_from(self, "last_seen")

    def delete_row(self):
        for element in self.handlers:
            element.delete()


@dataclasses.dataclass
class WorkersSection:
    workers: Dict[str, WorkerRow] = dataclasses.field(default_factory=lambda: defaultdict(WorkerRow))

    @ui.refreshable
    def draw_section(self):
        with ui.row().classes("h-max"), ui.card().classes("w-full"), ui.grid(columns=12):
            self.__draw_titles()
            for worker_row in self.workers.values():
                worker_row.draw_row()

    @staticmethod
    def __draw_titles():
        ui.label("Worker")
        ui.label("Agt CPU %")
        ui.label("Agt RSS (in MB)")
        ui.label("Processors CPU %")
        ui.label("Processors RSS (in MB)")
        ui.label("Queue Capacity")
        ui.label("Tasks Sent")
        ui.label("Tasks Queued")
        ui.label("Tasks Suspended")
        ui.label("Lag")
        ui.label("ITL")
        ui.label("Last Seen")



from typing import Dict, Generic, Set, TypeVar

KeyT = TypeVar("KeyT")
ValueT = TypeVar("ValueT")


class OneToManyDict(Generic[KeyT, ValueT]):
    def __init__(self):
        self._key_to_value_set: Dict[KeyT, Set[ValueT]] = dict()
        self._value_to_key: Dict[ValueT, KeyT] = dict()

    def __contains__(self, key) -> bool:
        return self.has_key(key)

    def keys(self):
        return self._key_to_value_set.keys()

    def values(self):
        return self._key_to_value_set.values()

    def items(self):
        return self._key_to_value_set.items()

    def add(self, key: KeyT, value: ValueT):
        if value in self._value_to_key and self._value_to_key[value] != key:
            raise ValueError("value has to be unique in OneToManyDict")

        self._value_to_key[value] = key

        if key not in self._key_to_value_set:
            self._key_to_value_set[key] = set()

        self._key_to_value_set[key].add(value)

    def has_key(self, key: KeyT) -> bool:
        return key in self._key_to_value_set

    def has_value(self, value: ValueT) -> bool:
        return value in self._value_to_key

    def get_key(self, value: ValueT) -> KeyT:
        if value not in self._value_to_key:
            raise ValueError(f"cannot find {value=} in OneToManyDict")

        return self._value_to_key[value]

    def get_values(self, key: KeyT) -> Set[ValueT]:
        if key not in self._key_to_value_set:
            raise ValueError(f"cannot find {key=} in OneToManyDict")

        return self._key_to_value_set[key]

    def remove_key(self, key: KeyT) -> Set[ValueT]:
        if key not in self._key_to_value_set:
            raise KeyError(f"cannot find {key=} in OneToManyDict")

        values = self._key_to_value_set.pop(key)
        for value in values:
            self._value_to_key.pop(value)

        return values

    def remove_value(self, value: ValueT) -> KeyT:
        if value not in self._value_to_key:
            raise ValueError(f"cannot find {value=} in OneToManyDict")

        key = self._value_to_key.pop(value)
        self._key_to_value_set[key].remove(value)
        if not self._key_to_value_set[key]:
            self._key_to_value_set.pop(key)

        return key



import logging
import socket
from typing import Optional, Tuple

from scaler.cluster.cluster import Cluster
from scaler.cluster.scheduler import SchedulerProcess
from scaler.io.config import (
    DEFAULT_CLIENT_TIMEOUT_SECONDS,
    DEFAULT_GARBAGE_COLLECT_INTERVAL_SECONDS,
    DEFAULT_HARD_PROCESSOR_SUSPEND,
    DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
    DEFAULT_IO_THREADS,
    DEFAULT_LOAD_BALANCE_SECONDS,
    DEFAULT_LOAD_BALANCE_TRIGGER_TIMES,
    DEFAULT_MAX_NUMBER_OF_TASKS_WAITING,
    DEFAULT_OBJECT_RETENTION_SECONDS,
    DEFAULT_PER_WORKER_QUEUE_SIZE,
    DEFAULT_TASK_TIMEOUT_SECONDS,
    DEFAULT_TRIM_MEMORY_THRESHOLD_BYTES,
    DEFAULT_WORKER_DEATH_TIMEOUT,
    DEFAULT_WORKER_TIMEOUT_SECONDS,
)
from scaler.utility.zmq_config import ZMQConfig


class SchedulerClusterCombo:
    def __init__(
        self,
        address: str,
        n_workers: int,
        worker_io_threads: int = DEFAULT_IO_THREADS,
        scheduler_io_threads: int = DEFAULT_IO_THREADS,
        max_number_of_tasks_waiting: int = DEFAULT_MAX_NUMBER_OF_TASKS_WAITING,
        heartbeat_interval_seconds: int = DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
        client_timeout_seconds: int = DEFAULT_CLIENT_TIMEOUT_SECONDS,
        worker_timeout_seconds: int = DEFAULT_WORKER_TIMEOUT_SECONDS,
        object_retention_seconds: int = DEFAULT_OBJECT_RETENTION_SECONDS,
        task_timeout_seconds: int = DEFAULT_TASK_TIMEOUT_SECONDS,
        death_timeout_seconds: int = DEFAULT_WORKER_DEATH_TIMEOUT,
        load_balance_seconds: int = DEFAULT_LOAD_BALANCE_SECONDS,
        load_balance_trigger_times=DEFAULT_LOAD_BALANCE_TRIGGER_TIMES,
        garbage_collect_interval_seconds: int = DEFAULT_GARBAGE_COLLECT_INTERVAL_SECONDS,
        trim_memory_threshold_bytes: int = DEFAULT_TRIM_MEMORY_THRESHOLD_BYTES,
        per_worker_queue_size: int = DEFAULT_PER_WORKER_QUEUE_SIZE,
        hard_processor_suspend: bool = DEFAULT_HARD_PROCESSOR_SUSPEND,
        protected: bool = True,
        event_loop: str = "builtin",
        logging_paths: Tuple[str, ...] = ("/dev/stdout",),
        logging_level: str = "INFO",
        logging_config_file: Optional[str] = None,
    ):
        self._cluster = Cluster(
            address=ZMQConfig.from_string(address),
            worker_io_threads=worker_io_threads,
            worker_names=[f"{socket.gethostname().split('.')[0]}_{i}" for i in range(n_workers)],
            heartbeat_interval_seconds=heartbeat_interval_seconds,
            task_timeout_seconds=task_timeout_seconds,
            death_timeout_seconds=death_timeout_seconds,
            garbage_collect_interval_seconds=garbage_collect_interval_seconds,
            trim_memory_threshold_bytes=trim_memory_threshold_bytes,
            hard_processor_suspend=hard_processor_suspend,
            event_loop=event_loop,
            logging_paths=logging_paths,
            logging_level=logging_level,
            logging_config_file=logging_config_file,
        )
        self._scheduler = SchedulerProcess(
            address=ZMQConfig.from_string(address),
            io_threads=scheduler_io_threads,
            max_number_of_tasks_waiting=max_number_of_tasks_waiting,
            per_worker_queue_size=per_worker_queue_size,
            client_timeout_seconds=client_timeout_seconds,
            worker_timeout_seconds=worker_timeout_seconds,
            object_retention_seconds=object_retention_seconds,
            load_balance_seconds=load_balance_seconds,
            load_balance_trigger_times=load_balance_trigger_times,
            protected=protected,
            event_loop=event_loop,
            logging_path=logging_paths,
            logging_config_file=logging_config_file,
        )

        self._cluster.start()
        self._scheduler.start()
        logging.info(f"{self.__get_prefix()} started")

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        logging.info(f"{self.__get_prefix()} shutdown")
        self._cluster.terminate()
        self._scheduler.terminate()
        self._cluster.join()
        self._scheduler.join()

    def __get_prefix(self):
        return f"{self.__class__.__name__}:"



import asyncio
import multiprocessing
from asyncio import AbstractEventLoop, Task
from typing import Any, Optional, Tuple

from scaler.scheduler.config import SchedulerConfig
from scaler.scheduler.scheduler import Scheduler, scheduler_main
from scaler.utility.event_loop import register_event_loop
from scaler.utility.logging.utility import setup_logger
from scaler.utility.zmq_config import ZMQConfig


class SchedulerProcess(multiprocessing.get_context("spawn").Process):  # type: ignore[misc]
    def __init__(
        self,
        address: ZMQConfig,
        io_threads: int,
        max_number_of_tasks_waiting: int,
        per_worker_queue_size: int,
        client_timeout_seconds: int,
        worker_timeout_seconds: int,
        object_retention_seconds: int,
        load_balance_seconds: int,
        load_balance_trigger_times: int,
        protected: bool,
        event_loop: str,
        logging_path: Tuple[str, ...],
        logging_config_file: Optional[str],
    ):
        multiprocessing.Process.__init__(self, name="Scheduler")
        self._scheduler_config = SchedulerConfig(
            event_loop=event_loop,
            address=address,
            io_threads=io_threads,
            max_number_of_tasks_waiting=max_number_of_tasks_waiting,
            per_worker_queue_size=per_worker_queue_size,
            client_timeout_seconds=client_timeout_seconds,
            worker_timeout_seconds=worker_timeout_seconds,
            object_retention_seconds=object_retention_seconds,
            load_balance_seconds=load_balance_seconds,
            load_balance_trigger_times=load_balance_trigger_times,
            protected=protected,
        )

        self._logging_path = logging_path
        self._logging_config_file = logging_config_file

        self._scheduler: Optional[Scheduler] = None
        self._loop: Optional[AbstractEventLoop] = None
        self._task: Optional[Task[Any]] = None

    def run(self) -> None:
        # scheduler have its own single process
        setup_logger(self._logging_path, self._logging_config_file)
        register_event_loop(self._scheduler_config.event_loop)

        self._loop = asyncio.get_event_loop()
        self._task = self._loop.create_task(scheduler_main(self._scheduler_config))

        self._loop.run_until_complete(self._task)



import random
import time
import unittest

from tests.utility import get_available_tcp_port, logging_test_name

from scaler import Client, SchedulerClusterCombo
from scaler.utility.logging.scoped_logger import ScopedLogger
from scaler.utility.logging.utility import setup_logger


def noop(sec: int):
    return sec * 1


def noop_sleep(sec: int):
    time.sleep(sec)
    return sec


def noop_memory(length: int):
    time.sleep(5)
    chunk_size = 1024
    chunks = []
    for i in range(length):
        chunks.append(bytearray(chunk_size))
    return len(chunks)


class TestUI(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)
        self.address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        self._workers = 10
        self.cluster = SchedulerClusterCombo(address=self.address, n_workers=self._workers, event_loop="builtin")

    def tearDown(self) -> None:
        self.cluster.shutdown()
        pass

    def test_noop(self):
        with Client(self.address) as client:
            tasks = [random.randint(2, 50) for _ in range(20)]
            with ScopedLogger(f"submit {len(tasks)} noop tasks"):
                futures = [client.submit(noop_sleep, i) for i in tasks]

            with ScopedLogger(f"gather {len(futures)} results"):
                results = [future.result() for future in futures]

            self.assertEqual(results, tasks)
            time.sleep(2)

            tasks = [random.randint(2, 50) for _ in range(20)]
            with ScopedLogger(f"submit {len(tasks)} noop tasks"):
                futures = [client.submit(noop_sleep, i) for i in tasks]

            with ScopedLogger(f"gather {len(futures)} results"):
                results = [future.result() for future in futures]

            self.assertEqual(results, tasks)

    def test_memory_usage(self):
        with Client(self.address) as client:
            tasks = [0, 1, 100, 1024, 1024 * 2, 1024 * 1024, 1024 * 1024 * 2, 1024 * 1024 * 4]
            results = []

            for task in tasks:
                with ScopedLogger(f"submit {task} memory task"):
                    future = client.submit(noop_memory, task)

                with ScopedLogger(f"gather {task} result"):
                    results.append(future.result())
                time.sleep(2)

            self.assertEqual(results, tasks)



import hashlib
import pickle
from typing import Tuple

SERIALIZER_INDICATOR = b"serializer"
POSTFIX_SERIALIZER = hashlib.md5(SERIALIZER_INDICATOR).digest()


def generate_object_id(identity: bytes, object_bytes: bytes) -> bytes:
    identity_hash = hashlib.md5(identity).digest()
    object_hash = hashlib.md5(object_bytes).digest()
    return identity_hash[:8] + object_hash


def split_object_id(object_id: bytes) -> Tuple[bytes, bytes]:
    return object_id[:7], object_id[8:]


def generate_serializer_object_id(source: bytes) -> bytes:
    return generate_object_id(source, SERIALIZER_INDICATOR)


def is_object_id_serializer(object_id: bytes) -> bool:
    return object_id.endswith(POSTFIX_SERIALIZER)


def serialize_failure(exp: Exception) -> bytes:
    return pickle.dumps(exp, protocol=pickle.HIGHEST_PROTOCOL)


def deserialize_failure(result: bytes) -> Exception:
    return pickle.loads(result)



from typing import Dict, Optional, Set

from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.common import TaskStatus
from scaler.protocol.python.message import Task, TaskCancel, TaskResult
from scaler.utility.metadata.task_flags import retrieve_task_flags_from_task
from scaler.utility.mixins import Looper
from scaler.utility.queues.async_sorted_priority_queue import AsyncSortedPriorityQueue
from scaler.worker.agent.mixins import ProcessorManager, TaskManager

_SUSPENDED_TASKS_PRIORITY = 1
_QUEUED_TASKS_PRIORITY = 2


class VanillaTaskManager(Looper, TaskManager):
    def __init__(self, task_timeout_seconds: int):
        self._task_timeout_seconds = task_timeout_seconds

        self._queued_task_id_to_task: Dict[bytes, Task] = dict()

        # Queued tasks are sorted first by task's priorities, then suspended tasks are prioritized over non yet started
        # tasks. Finally the sorted queue ensure we execute the oldest tasks first.
        #
        # For example, if we receive these tasks in this order:
        #   1. Task(priority=0) [suspended]
        #   2. Task(priority=3) [suspended]
        #   3. Task(priority=3)
        #   4. Task(priority=0)
        #
        # We want to execute the tasks in this order: 2-3-1-4.
        self._queued_task_ids = AsyncSortedPriorityQueue()

        self._processing_task_ids: Set[bytes] = set()  # Tasks associated with a processor, including suspended tasks

        self._connector_external: Optional[AsyncConnector] = None
        self._processor_manager: Optional[ProcessorManager] = None

    def register(self, connector: AsyncConnector, processor_manager: ProcessorManager):
        self._connector_external = connector
        self._processor_manager = processor_manager

    async def on_task_new(self, task: Task):
        self.__enqueue_task(task, is_suspended=False)

        await self.__suspend_if_priority_is_higher(task)

    async def on_cancel_task(self, task_cancel: TaskCancel):
        task_id = task_cancel.task_id

        task_not_found = task_id not in self._processing_task_ids and task_id not in self._queued_task_id_to_task
        task_is_processing = task_id in self._processing_task_ids

        if task_not_found or (task_is_processing and not task_cancel.flags.force):
            result = TaskResult.new_msg(task_id, TaskStatus.NotFound)
            await self._connector_external.send(result)
            return

        # A suspended task will be both processing AND queued

        if task_cancel.task_id in self._queued_task_id_to_task:
            canceled_task = self._queued_task_id_to_task.pop(task_cancel.task_id)
            self._queued_task_ids.remove(task_cancel.task_id)

        if task_is_processing:
            self._processing_task_ids.remove(task_cancel.task_id)
            canceled_task = await self._processor_manager.on_cancel_task(task_cancel.task_id)

        assert canceled_task is not None

        payload = [canceled_task.get_message().to_bytes()] if task_cancel.flags.retrieve_task_object else []
        result = TaskResult.new_msg(task_id=task_id, status=TaskStatus.Canceled, metadata=b"", results=payload)
        await self._connector_external.send(result)

    async def on_task_result(self, result: TaskResult):
        if result.task_id in self._queued_task_id_to_task:
            # Finishing a queued task might happen if a task ended during the suspension process.
            self._queued_task_id_to_task.pop(result.task_id)
            self._queued_task_ids.remove(result.task_id)

        self._processing_task_ids.remove(result.task_id)

        await self._connector_external.send(result)

    async def routine(self):
        await self.__processing_task()

    def get_queued_size(self):
        return self._queued_task_ids.qsize()

    async def __processing_task(self):
        await self._processor_manager.wait_until_can_accept_task()

        _, task_id = await self._queued_task_ids.get()
        task = self._queued_task_id_to_task.pop(task_id)

        if task_id not in self._processing_task_ids:
            self._processing_task_ids.add(task_id)
            await self._processor_manager.on_task(task)
        else:
            self._processor_manager.on_resume_task(task_id)

    async def __suspend_if_priority_is_higher(self, new_task: Task):
        current_task = self._processor_manager.current_task()

        if current_task is None:
            return

        new_task_priority = self.__get_task_priority(new_task)
        current_task_priority = self.__get_task_priority(current_task)

        if new_task_priority <= current_task_priority:
            return

        self.__enqueue_task(current_task, is_suspended=True)

        await self._processor_manager.on_suspend_task(current_task.task_id)

    def __enqueue_task(self, task: Task, is_suspended: bool):
        task_priority = self.__get_task_priority(task)

        # Higher-priority tasks have an higher priority value. But as the queue is sorted by increasing order, we negate
        # the inserted value so they will be at the head of the queue.
        if is_suspended:
            queue_priority = (-task_priority, _SUSPENDED_TASKS_PRIORITY)
        else:
            queue_priority = (-task_priority, _QUEUED_TASKS_PRIORITY)

        self._queued_task_ids.put_nowait((queue_priority, task.task_id))
        self._queued_task_id_to_task[task.task_id] = task

    @staticmethod
    def __get_task_priority(task: Task) -> int:
        priority = retrieve_task_flags_from_task(task).priority

        if priority < 0:
            raise ValueError(f"invalid task priority, must be positive or zero, got {priority}")

        return priority



import dataclasses


@dataclasses.dataclass
class ObjectReference:
    name: bytes
    object_id: bytes
    size: int

    def __repr__(self):
        return f"ScalerReference(name={self.name!r}, id={self.object_id!r}, size={self.size})"

    def __hash__(self):
        return hash(self.object_id)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ObjectReference):
            return NotImplemented

        return self.object_id == other.object_id

    def __ne__(self, other):
        return not self.__eq__(other)



import logging
import time
from typing import Dict, Optional, Set, Tuple

from scaler.io.async_binder import AsyncBinder
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.message import (
    ClientDisconnect,
    ClientHeartbeat,
    ClientHeartbeatEcho,
    ClientShutdownResponse,
    TaskCancel,
)
from scaler.protocol.python.status import ClientManagerStatus
from scaler.scheduler.mixins import ClientManager, ObjectManager, TaskManager, WorkerManager
from scaler.utility.exceptions import ClientShutdownException
from scaler.utility.mixins import Looper, Reporter
from scaler.utility.one_to_many_dict import OneToManyDict


class VanillaClientManager(ClientManager, Looper, Reporter):
    def __init__(self, client_timeout_seconds: int, protected: bool):
        self._client_timeout_seconds = client_timeout_seconds
        self._protected = protected
        self._client_to_task_ids: OneToManyDict[bytes, bytes] = OneToManyDict()

        self._binder: Optional[AsyncBinder] = None
        self._binder_monitor: Optional[AsyncConnector] = None
        self._object_manager: Optional[ObjectManager] = None
        self._task_manager: Optional[TaskManager] = None
        self._worker_manager: Optional[WorkerManager] = None

        self._client_last_seen: Dict[bytes, Tuple[float, ClientHeartbeat]] = dict()

    def register(
        self,
        binder: AsyncBinder,
        binder_monitor: AsyncConnector,
        object_manager: ObjectManager,
        task_manager: TaskManager,
        worker_manager: WorkerManager,
    ):
        self._binder = binder
        self._binder_monitor = binder_monitor
        self._object_manager = object_manager
        self._task_manager = task_manager
        self._worker_manager = worker_manager

    def get_client_task_ids(self, client: bytes) -> Set[bytes]:
        return self._client_to_task_ids.get_values(client)

    def has_client_id(self, client_id: bytes) -> bool:
        return client_id in self._client_last_seen

    def get_client_id(self, task_id: bytes) -> Optional[bytes]:
        return self._client_to_task_ids.get_key(task_id)

    def on_task_begin(self, client: bytes, task_id: bytes):
        self._client_to_task_ids.add(client, task_id)

    def on_task_finish(self, task_id: bytes) -> bytes:
        return self._client_to_task_ids.remove_value(task_id)

    async def on_heartbeat(self, client: bytes, info: ClientHeartbeat):
        await self._binder.send(client, ClientHeartbeatEcho.new_msg())
        if client not in self._client_last_seen:
            logging.info(f"client {client!r} connected")

        self._client_last_seen[client] = (time.time(), info)

    async def on_client_disconnect(self, client: bytes, request: ClientDisconnect):
        if request.disconnect_type == ClientDisconnect.DisconnectType.Disconnect:
            await self.__on_client_disconnect(client)
            return

        if self._protected:
            logging.warning("cannot shutdown clusters as scheduler is running in protected mode")
            accepted = False
        else:
            logging.info(f"shutdown scheduler and all clusters as received signal from {client=!r}")
            accepted = True

        await self._binder.send(client, ClientShutdownResponse.new_msg(accepted=accepted))

        if self._protected:
            return

        await self._worker_manager.on_client_shutdown(client)

        raise ClientShutdownException(f"received client shutdown from {client!r}, quiting")

    async def routine(self):
        await self.__routine_cleanup_clients()

    def get_status(self) -> ClientManagerStatus:
        return ClientManagerStatus.new_msg(
            {client.decode(): len(task_ids) for client, task_ids in self._client_to_task_ids.items()}
        )

    async def __routine_cleanup_clients(self):
        now = time.time()
        dead_clients = {
            client
            for client, (last_seen, info) in self._client_last_seen.items()
            if now - last_seen > self._client_timeout_seconds
        }

        for client in dead_clients:
            await self.__on_client_disconnect(client)

    async def __on_client_disconnect(self, client_id: bytes):
        logging.info(f"client {client_id!r} disconnected")
        if client_id in self._client_last_seen:
            self._client_last_seen.pop(client_id)

        await self.__cancel_tasks(client_id)
        self._object_manager.clean_client(client_id)

    async def __cancel_tasks(self, client: bytes):
        if client not in self._client_to_task_ids.keys():
            return

        tasks = self._client_to_task_ids.get_values(client).copy()
        for task in tasks:
            await self._task_manager.on_task_cancel(client, TaskCancel.new_msg(task))



from .about import __version__
from .client.client import Client, ScalerFuture
from .client.serializer.mixins import Serializer
from .cluster.cluster import Cluster
from .cluster.combo import SchedulerClusterCombo
from .cluster.scheduler import Scheduler

assert isinstance(__version__, str)
assert isinstance(Client, type)
assert isinstance(ScalerFuture, type)
assert isinstance(Scheduler, type)
assert isinstance(Cluster, type)
assert isinstance(SchedulerClusterCombo, type)
assert isinstance(Serializer, type)



import logging
import time
from typing import Dict, List, Optional, Set, Tuple

from scaler.io.async_binder import AsyncBinder
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.common import TaskStatus
from scaler.protocol.python.message import (
    ClientDisconnect,
    DisconnectRequest,
    DisconnectResponse,
    StateBalanceAdvice,
    StateWorker,
    Task,
    TaskCancel,
    TaskResult,
    WorkerHeartbeat,
    WorkerHeartbeatEcho,
)
from scaler.protocol.python.status import ProcessorStatus, Resource, WorkerManagerStatus, WorkerStatus
from scaler.scheduler.allocators.queued import QueuedAllocator
from scaler.scheduler.mixins import TaskManager, WorkerManager
from scaler.utility.mixins import Looper, Reporter

UINT8_MAX = 2**8 - 1


class VanillaWorkerManager(WorkerManager, Looper, Reporter):
    def __init__(
        self,
        per_worker_queue_size: int,
        timeout_seconds: int,
        load_balance_seconds: int,
        load_balance_trigger_times: int,
    ):
        self._timeout_seconds = timeout_seconds
        self._load_balance_seconds = load_balance_seconds
        self._load_balance_trigger_times = load_balance_trigger_times

        self._binder: Optional[AsyncBinder] = None
        self._binder_monitor: Optional[AsyncConnector] = None
        self._task_manager: Optional[TaskManager] = None

        self._worker_alive_since: Dict[bytes, Tuple[float, WorkerHeartbeat]] = dict()
        self._allocator = QueuedAllocator(per_worker_queue_size)

        self._last_balance_advice: Dict[bytes, List[bytes]] = dict()
        self._load_balance_advice_same_count = 0

    def register(self, binder: AsyncBinder, binder_monitor: AsyncConnector, task_manager: TaskManager):
        self._binder = binder
        self._binder_monitor = binder_monitor
        self._task_manager = task_manager

    async def assign_task_to_worker(self, task: Task) -> bool:
        worker = await self._allocator.assign_task(task.task_id)
        if worker is None:
            return False

        # send to worker
        await self._binder.send(worker, task)
        return True

    async def on_task_cancel(self, task_cancel: TaskCancel):
        worker = self._allocator.remove_task(task_cancel.task_id)
        if worker is None:
            logging.error(f"cannot find task_id={task_cancel.task_id.hex()} in task workers")
            return

        await self._binder.send(worker, TaskCancel.new_msg(task_cancel.task_id))

    async def on_task_result(self, task_result: TaskResult):
        worker = self._allocator.remove_task(task_result.task_id)

        if task_result.status in {TaskStatus.Canceled, TaskStatus.NotFound}:
            if worker is not None:
                # The worker canceled the task, but the scheduler still had it queued. Re-route the task to another
                # worker.
                await self.__reroute_tasks([task_result.task_id])
            else:
                await self._task_manager.on_task_done(task_result)

            return

        if worker is None:
            logging.error(
                f"received unknown task result for task_id={task_result.task_id.hex()}, status={task_result.status} "
                f"might due to worker get disconnected or canceled"
            )
            return

        await self._task_manager.on_task_done(task_result)

    async def on_heartbeat(self, worker: bytes, info: WorkerHeartbeat):
        if await self._allocator.add_worker(worker):
            logging.info(f"worker {worker!r} connected")
            await self._binder_monitor.send(StateWorker.new_msg(worker, b"connected"))

        self._worker_alive_since[worker] = (time.time(), info)
        await self._binder.send(worker, WorkerHeartbeatEcho.new_msg())

    async def on_client_shutdown(self, client: bytes):
        for worker in self._allocator.get_worker_ids():
            await self.__shutdown_worker(worker)

    async def on_disconnect(self, source: bytes, request: DisconnectRequest):
        await self.__disconnect_worker(request.worker)
        await self._binder.send(source, DisconnectResponse.new_msg(request.worker))

    async def routine(self):
        await self.__balance_request()
        await self.__clean_workers()

    def get_status(self) -> WorkerManagerStatus:
        worker_to_task_numbers = self._allocator.statistics()
        return WorkerManagerStatus.new_msg(
            [
                self.__worker_status_from_heartbeat(worker, worker_to_task_numbers[worker], last, info)
                for worker, (last, info) in self._worker_alive_since.items()
            ]
        )

    @staticmethod
    def __worker_status_from_heartbeat(
        worker: bytes, worker_task_numbers: Dict, last: float, info: WorkerHeartbeat
    ) -> WorkerStatus:
        current_processor = next((p for p in info.processors if not p.suspended), None)
        suspended = min(len([p for p in info.processors if p.suspended]), UINT8_MAX)
        last_s = min(int(time.time() - last), UINT8_MAX)

        if current_processor:
            debug_info = f"{int(current_processor.initialized)}{int(current_processor.has_task)}{int(info.task_lock)}"
        else:
            debug_info = f"00{int(info.task_lock)}"

        return WorkerStatus.new_msg(
            worker_id=worker,
            agent=info.agent,
            rss_free=info.rss_free,
            free=worker_task_numbers["free"],
            sent=worker_task_numbers["sent"],
            queued=info.queued_tasks,
            suspended=suspended,
            lag_us=info.latency_us,
            last_s=last_s,
            itl=debug_info,
            processor_statuses=[
                ProcessorStatus.new_msg(
                    pid=p.pid,
                    initialized=p.initialized,
                    has_task=p.has_task,
                    suspended=p.suspended,
                    resource=Resource.new_msg(p.resource.cpu, p.resource.rss),
                )
                for p in info.processors
            ],
        )

    def has_available_worker(self) -> bool:
        return self._allocator.has_available_worker()

    def get_worker_by_task_id(self, task_id: bytes) -> bytes:
        return self._allocator.get_worker_by_task_id(task_id)

    def get_worker_ids(self) -> Set[bytes]:
        return self._allocator.get_worker_ids()

    async def __balance_request(self):
        if self._load_balance_seconds <= 0:
            return

        current_advice = self._allocator.balance()
        if self._last_balance_advice == current_advice:
            self._load_balance_advice_same_count += 1
        else:
            self._last_balance_advice = current_advice
            self._load_balance_advice_same_count = 0

        if 0 < self._load_balance_advice_same_count < self._load_balance_trigger_times:
            return

        await self.__do_balance(current_advice)

    async def __do_balance(self, current_advice: Dict[bytes, List[bytes]]):
        if not current_advice:
            return

        worker_to_num_tasks = {worker: len(task_ids) for worker, task_ids in current_advice.items()}
        logging.info(f"balancing task: {worker_to_num_tasks}")
        for worker, task_ids in current_advice.items():
            await self._binder_monitor.send(StateBalanceAdvice.new_msg(worker, task_ids))

        task_cancel_flags = TaskCancel.TaskCancelFlags(force=True, retrieve_task_object=False)

        self._last_balance_advice = current_advice
        for worker, task_ids in current_advice.items():
            for task_id in task_ids:
                await self._binder.send(worker, TaskCancel.new_msg(task_id=task_id, flags=task_cancel_flags))

    async def __clean_workers(self):
        now = time.time()
        dead_workers = [
            dead_worker
            for dead_worker, (alive_since, info) in self._worker_alive_since.items()
            if now - alive_since > self._timeout_seconds
        ]
        for dead_worker in dead_workers:
            await self.__disconnect_worker(dead_worker)

    async def __reroute_tasks(self, task_ids: List[bytes]):
        for task_id in task_ids:
            await self._task_manager.on_task_reroute(task_id)

    async def __disconnect_worker(self, worker: bytes):
        """return True if disconnect worker success"""
        if worker not in self._worker_alive_since:
            return

        logging.info(f"worker {worker!r} disconnected")
        await self._binder_monitor.send(StateWorker.new_msg(worker, b"disconnected"))
        self._worker_alive_since.pop(worker)

        task_ids = self._allocator.remove_worker(worker)
        if not task_ids:
            return

        logging.info(f"rerouting {len(task_ids)} tasks")
        await self.__reroute_tasks(task_ids)

    async def __shutdown_worker(self, worker: bytes):
        await self._binder.send(worker, ClientDisconnect.new_msg(ClientDisconnect.DisconnectType.Shutdown))
        await self.__disconnect_worker(worker)



import time
from typing import Optional

import psutil

from scaler.client.agent.mixins import HeartbeatManager
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.message import ClientHeartbeat, ClientHeartbeatEcho
from scaler.protocol.python.status import Resource
from scaler.utility.mixins import Looper


class ClientHeartbeatManager(Looper, HeartbeatManager):
    def __init__(self, death_timeout_seconds: int):
        self._death_timeout_seconds = death_timeout_seconds

        self._process = psutil.Process()

        self._last_scheduler_contact = time.time()
        self._start_timestamp_ns = 0
        self._latency_us = 0
        self._connected = False

        self._connector_external: Optional[AsyncConnector] = None

    def register(self, connector_external: AsyncConnector):
        self._connector_external = connector_external

    async def send_heartbeat(self):
        await self._connector_external.send(
            ClientHeartbeat.new_msg(
                Resource.new_msg(int(self._process.cpu_percent() * 10), self._process.memory_info().rss),
                self._latency_us,
            )
        )

    async def on_heartbeat_echo(self, heartbeat: ClientHeartbeatEcho):
        if not self._connected:
            self._connected = True

        self._last_scheduler_contact = time.time()
        if self._start_timestamp_ns == 0:
            # not handling echo if we didn't send out heartbeat
            return

        self._latency_us = int(((time.time_ns() - self._start_timestamp_ns) / 2) // 1_000)
        self._start_timestamp_ns = 0

    async def routine(self):
        if time.time() - self._last_scheduler_contact > self._death_timeout_seconds:
            raise TimeoutError(
                f"Timeout when connecting to scheduler {self._connector_external.address} "
                f"in {self._death_timeout_seconds} seconds"
            )

        if self._start_timestamp_ns != 0:
            # already sent heartbeat, expecting heartbeat echo, so not sending
            return

        await self.send_heartbeat()
        self._start_timestamp_ns = time.time_ns()



import dataclasses
import logging
import time
from typing import Dict, Optional

import psutil

from scaler.utility.metadata.profile_result import ProfileResult
from scaler.utility.mixins import Looper
from scaler.worker.agent.mixins import ProfilingManager


@dataclasses.dataclass
class _ProcessProfiler:
    process: psutil.Process

    current_task_id: Optional[bytes] = None

    start_time: Optional[float] = None
    start_cpu_time: Optional[float] = None
    init_memory_rss: Optional[int] = None
    peak_memory_rss: Optional[int] = None


class VanillaProfilingManager(ProfilingManager, Looper):
    def __init__(self):
        self._process_profiler_by_pid: Dict[int, _ProcessProfiler] = {}

    def on_process_start(self, pid: int):
        if pid in self._process_profiler_by_pid:
            raise ValueError(f"process {pid=} is already registered.")

        self._process_profiler_by_pid[pid] = _ProcessProfiler(psutil.Process(pid))

    def on_process_end(self, pid: int):
        if pid not in self._process_profiler_by_pid:
            raise ValueError(f"process {pid=} is not registered.")

        self._process_profiler_by_pid.pop(pid)

    def on_task_start(self, pid: int, task_id: bytes):
        process_profiler = self._process_profiler_by_pid.get(pid)

        if process_profiler is None:
            raise ValueError(f"process {pid=} is not registered.")

        process_profiler.current_task_id = task_id

        process = process_profiler.process

        process_profiler.start_time = self.__process_time()
        process_profiler.start_cpu_time = self.__process_cpu_time(process)
        process_profiler.init_memory_rss = self.__process_memory_rss(process)
        process_profiler.peak_memory_rss = process_profiler.init_memory_rss

    def on_task_end(self, pid: int, task_id: bytes) -> ProfileResult:
        process_profiler = self._process_profiler_by_pid.get(pid)

        if process_profiler is None:
            raise ValueError(f"process {pid=} is not registered.")

        if task_id != process_profiler.current_task_id:
            raise ValueError(f"task {task_id=!r} is not the current task task_id={process_profiler.current_task_id!r}.")

        assert process_profiler.start_time is not None
        assert process_profiler.init_memory_rss is not None
        assert process_profiler.peak_memory_rss is not None

        process = process_profiler.process

        time_delta = self.__process_time() - process_profiler.start_time

        try:
            cpu_time_delta = self.__process_cpu_time(process) - process_profiler.start_cpu_time
        except psutil.ZombieProcess:
            logging.warning(f"profiling zombie process: {pid=}")
            cpu_time_delta = 0

        memory_delta = process_profiler.peak_memory_rss - process_profiler.init_memory_rss

        process_profiler.current_task_id = None
        process_profiler.init_memory_rss = None
        process_profiler.peak_memory_rss = None

        return ProfileResult(time_delta, memory_delta, cpu_time_delta)

    async def routine(self):
        for process_profiler in self._process_profiler_by_pid.values():
            if process_profiler.current_task_id is not None:
                try:
                    process_profiler.peak_memory_rss = max(
                        process_profiler.peak_memory_rss, self.__process_memory_rss(process_profiler.process)
                    )
                except psutil.ZombieProcess:
                    logging.warning(f"profiling zombie process: pid={process_profiler.process.pid}")

    @staticmethod
    def __process_time():
        return time.monotonic()

    @staticmethod
    def __process_cpu_time(process: psutil.Process) -> float:
        cpu_times = process.cpu_times()
        return cpu_times.user + cpu_times.system

    @staticmethod
    def __process_memory_rss(process: psutil.Process) -> int:
        return process.memory_info().rss



import unittest

from tests.utility import logging_test_name

from scaler import Client, SchedulerClusterCombo
from scaler.utility.logging.utility import setup_logger

N_TASKS = 30
N_WORKERS = 3
assert N_TASKS >= N_WORKERS


class TestNestedTask(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)
        self.address = "tcp://127.0.0.1:23456"
        self.cluster = SchedulerClusterCombo(address=self.address, n_workers=N_WORKERS, event_loop="builtin")

    def tearDown(self) -> None:
        self.cluster.shutdown()

    def test_nested_task_arg_client(self) -> None:
        with Client(self.address) as client:
            result = client.submit(parent_task_arg_client, client).result()
            self.assertEqual(result, sum(nested_task(v) for v in range(0, N_TASKS)))

    def test_recursive_task(self) -> None:
        with Client(self.address) as client:
            result = client.submit(factorial, client, 10).result()
            self.assertEqual(result, 3_628_800)

    def test_multiple_recursive_task(self) -> None:
        with Client(self.address) as client:
            result = client.submit(fibonacci, client, 8).result()
            self.assertEqual(result, 21)


def parent_task_arg_client(client: Client) -> int:
    iterable = [(v,) for v in range(0, N_TASKS)]
    return sum(client.map(nested_task, iterable))


def nested_task(value: int) -> int:
    return value**2


def factorial(client: Client, value: int) -> int:
    if value == 0:
        return 1
    else:
        return value * client.submit(factorial, client, value - 1).result()


def fibonacci(client: Client, n: int):
    if n == 0:
        return 0
    elif n == 1:
        return 1
    else:
        a = client.submit(fibonacci, client, n - 1)
        b = client.submit(fibonacci, client, n - 2)
        return a.result() + b.result()



import time
import unittest

from tests.utility import get_available_tcp_port, logging_test_name

from scaler import Client, SchedulerClusterCombo
from scaler.utility.logging.utility import setup_logger


class TestProtected(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)

    def test_protected_true(self) -> None:
        address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        cluster = SchedulerClusterCombo(
            address=address, n_workers=2, per_worker_queue_size=2, event_loop="builtin", protected=True
        )
        print("wait for 3 seconds")
        time.sleep(3)

        with self.assertRaises(ValueError):
            client = Client(address=address)
            client.shutdown()

        time.sleep(3)

        cluster.shutdown()

    def test_protected_false(self) -> None:
        address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        cluster = SchedulerClusterCombo(
            address=address, n_workers=2, per_worker_queue_size=2, event_loop="builtin", protected=False
        )
        print("wait for 3 seconds")
        time.sleep(3)

        client = Client(address=address)
        client.shutdown()

        # wait scheduler received shutdown instruction and kill all workers, then call cluster.shutdown
        time.sleep(3)
        print("finish slept 3 seconds")
        cluster.shutdown()



import argparse
import socket

from scaler.cluster.cluster import Cluster
from scaler.io.config import (
    DEFAULT_GARBAGE_COLLECT_INTERVAL_SECONDS,
    DEFAULT_HARD_PROCESSOR_SUSPEND,
    DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
    DEFAULT_IO_THREADS,
    DEFAULT_NUMBER_OF_WORKER,
    DEFAULT_TASK_TIMEOUT_SECONDS,
    DEFAULT_TRIM_MEMORY_THRESHOLD_BYTES,
    DEFAULT_WORKER_DEATH_TIMEOUT,
)
from scaler.utility.event_loop import EventLoopType, register_event_loop
from scaler.utility.zmq_config import ZMQConfig


def get_args():
    parser = argparse.ArgumentParser(
        "standalone compute cluster", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--num-of-workers", "-n", type=int, default=DEFAULT_NUMBER_OF_WORKER, help="number of workers in cluster"
    )
    parser.add_argument(
        "--worker-names",
        "-wn",
        type=str,
        default=None,
        help="worker names to replace default worker names (host names), separate by comma",
    )
    parser.add_argument(
        "--heartbeat-interval",
        "-hi",
        type=int,
        default=DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
        help="number of seconds to send heartbeat interval",
    )
    parser.add_argument(
        "--task-timeout-seconds",
        "-tts",
        type=int,
        default=DEFAULT_TASK_TIMEOUT_SECONDS,
        help="number of seconds task treat as timeout and return an exception",
    )
    parser.add_argument(
        "--garbage-collect-interval-seconds",
        "-gc",
        type=int,
        default=DEFAULT_GARBAGE_COLLECT_INTERVAL_SECONDS,
        help="garbage collect interval seconds",
    )
    parser.add_argument(
        "--death-timeout-seconds", "-ds", type=int, default=DEFAULT_WORKER_DEATH_TIMEOUT, help="death timeout seconds"
    )
    parser.add_argument(
        "--trim-memory-threshold-bytes",
        "-tm",
        type=int,
        default=DEFAULT_TRIM_MEMORY_THRESHOLD_BYTES,
        help="number of bytes threshold to enable libc to trim memory",
    )
    parser.add_argument(
        "--event-loop", "-el", default="builtin", choices=EventLoopType.allowed_types(), help="select event loop type"
    )
    parser.add_argument(
        "--io-threads", "-it", default=DEFAULT_IO_THREADS, help="specify number of io threads per worker"
    )
    parser.add_argument(
        "--hard-processor-suspend",
        "-hps",
        action="store_true",
        default=DEFAULT_HARD_PROCESSOR_SUSPEND,
        help=(
            "When set, suspends worker processors using the SIGTSTP signal instead of a synchronization event, "
            "fully halting computation on suspended tasks. Note that this may cause some tasks to fail if they "
            "do not support being paused at the OS level (e.g. tasks requiring active network connections)."
        ),
    )
    parser.add_argument(
        "--log-hub-address", "-la", default=None, type=ZMQConfig.from_string, help="address for Worker send logs"
    )
    parser.add_argument(
        "--logging-paths",
        "-lp",
        nargs="*",
        type=str,
        default=("/dev/stdout",),
        help='specify where cluster log should logged to, it can be multiple paths, "/dev/stdout" is default for '
        "standard output, each worker will have its own log file with process id appended to the path",
    )
    parser.add_argument(
        "--logging-level",
        "-ll",
        type=str,
        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        default="INFO",
        help="specify the logging level",
    )
    parser.add_argument(
        "--logging-config-file",
        type=str,
        default=None,
        help="use standard python the .conf file the specify python logging file configuration format, this will "
        "bypass --logging-paths and --logging-level at the same time, and this will not work on per worker logging",
    )
    parser.add_argument("address", type=ZMQConfig.from_string, help="scheduler address to connect to")
    return parser.parse_args()


def main():
    args = get_args()
    register_event_loop(args.event_loop)

    if args.worker_names is None:
        worker_names = [f"{socket.gethostname().split('.')[0]}" for _ in range(args.num_of_workers)]
    else:
        worker_names = args.worker_names.split(",")
        if len(worker_names) != args.num_of_workers:
            raise ValueError(
                f"number of worker names ({len(args.worker_names)}) must match number of workers "
                f"({args.num_of_workers})"
            )

    cluster = Cluster(
        address=args.address,
        worker_names=worker_names,
        heartbeat_interval_seconds=args.heartbeat_interval,
        task_timeout_seconds=args.task_timeout_seconds,
        garbage_collect_interval_seconds=args.garbage_collect_interval_seconds,
        trim_memory_threshold_bytes=args.trim_memory_threshold_bytes,
        death_timeout_seconds=args.death_timeout_seconds,
        hard_processor_suspend=args.hard_processor_suspend,
        event_loop=args.event_loop,
        worker_io_threads=args.io_threads,
        logging_paths=args.logging_paths,
        logging_level=args.logging_level,
        logging_config_file=args.logging_config_file,
    )
    cluster.run()



import logging
import os
import uuid
from collections import defaultdict
from typing import Awaitable, Callable, Dict, List, Optional

import zmq.asyncio
from zmq import Frame

from scaler.io.utility import deserialize, serialize
from scaler.protocol.python.mixins import Message
from scaler.protocol.python.status import BinderStatus
from scaler.utility.mixins import Looper, Reporter
from scaler.utility.zmq_config import ZMQConfig


class AsyncBinder(Looper, Reporter):
    def __init__(self, context: zmq.asyncio.Context, name: str, address: ZMQConfig, identity: Optional[bytes] = None):
        self._address = address

        if identity is None:
            identity = f"{os.getpid()}|{name}|{uuid.uuid4()}".encode()
        self._identity = identity

        self._context = context
        self._socket = self._context.socket(zmq.ROUTER)
        self.__set_socket_options()
        self._socket.bind(self._address.to_address())

        self._callback: Optional[Callable[[bytes, Message], Awaitable[None]]] = None

        self._received: Dict[str, int] = defaultdict(lambda: 0)
        self._sent: Dict[str, int] = defaultdict(lambda: 0)

    def destroy(self):
        self._context.destroy(linger=0)

    def register(self, callback: Callable[[bytes, Message], Awaitable[None]]):
        self._callback = callback

    async def routine(self):
        frames: List[Frame] = await self._socket.recv_multipart(copy=False)
        if not self.__is_valid_message(frames):
            return

        source, payload = frames
        message: Optional[Message] = deserialize(payload.bytes)
        if message is None:
            logging.error(f"received unknown message from {source.bytes!r}: {payload!r}")
            return

        self.__count_received(message.__class__.__name__)
        await self._callback(source.bytes, message)

    async def send(self, to: bytes, message: Message):
        self.__count_sent(message.__class__.__name__)
        await self._socket.send_multipart([to, serialize(message)], copy=False)

    def get_status(self) -> BinderStatus:
        return BinderStatus.new_msg(received=self._received, sent=self._sent)

    def __set_socket_options(self):
        self._socket.setsockopt(zmq.IDENTITY, self._identity)
        self._socket.setsockopt(zmq.SNDHWM, 0)
        self._socket.setsockopt(zmq.RCVHWM, 0)

    def __is_valid_message(self, frames: List[Frame]) -> bool:
        if len(frames) < 2:
            logging.error(f"{self.__get_prefix()} received unexpected frames {frames}")
            return False

        return True

    def __count_received(self, message_type: str):
        self._received[message_type] += 1

    def __count_sent(self, message_type: str):
        self._sent[message_type] += 1

    def __get_prefix(self):
        return f"{self.__class__.__name__}[{self._identity.decode()}]:"



import logging

try:
    from scaler.utility.graph.topological_sorter_graphblas import TopologicalSorter

    logging.info("using GraphBLAS for calculate graph")
except ImportError as e:
    assert isinstance(e, Exception)
    from graphlib import TopologicalSorter  # type: ignore[assignment, no-redef]

    assert isinstance(TopologicalSorter, object)



import asyncio
import logging
import threading
from typing import Optional

import zmq.asyncio

from scaler.client.agent.disconnect_manager import ClientDisconnectManager
from scaler.client.agent.future_manager import ClientFutureManager
from scaler.client.agent.heartbeat_manager import ClientHeartbeatManager
from scaler.client.agent.object_manager import ClientObjectManager
from scaler.client.agent.task_manager import ClientTaskManager
from scaler.client.serializer.mixins import Serializer
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.message import (
    ClientDisconnect,
    ClientHeartbeatEcho,
    ClientShutdownResponse,
    GraphTask,
    GraphTaskCancel,
    ObjectInstruction,
    ObjectRequest,
    ObjectResponse,
    Task,
    TaskCancel,
    TaskResult,
)
from scaler.protocol.python.mixins import Message
from scaler.utility.event_loop import create_async_loop_routine
from scaler.utility.exceptions import ClientCancelledException, ClientQuitException, ClientShutdownException
from scaler.utility.zmq_config import ZMQConfig


class ClientAgent(threading.Thread):
    def __init__(
        self,
        identity: bytes,
        client_agent_address: ZMQConfig,
        scheduler_address: ZMQConfig,
        context: zmq.Context,
        future_manager: ClientFutureManager,
        stop_event: threading.Event,
        timeout_seconds: int,
        heartbeat_interval_seconds: int,
        serializer: Serializer,
    ):
        threading.Thread.__init__(self, daemon=True)

        self._stop_event = stop_event
        self._timeout_seconds = timeout_seconds
        self._heartbeat_interval_seconds = heartbeat_interval_seconds
        self._serializer = serializer

        self._identity = identity
        self._client_agent_address = client_agent_address
        self._scheduler_address = scheduler_address
        self._context = context

        self._future_manager = future_manager

        self._connector_internal = AsyncConnector(
            context=zmq.asyncio.Context.shadow(self._context),
            name="client_agent_internal",
            socket_type=zmq.PAIR,
            bind_or_connect="bind",
            address=self._client_agent_address,
            callback=self.__on_receive_from_client,
            identity=None,
        )
        self._connector_external = AsyncConnector(
            context=zmq.asyncio.Context.shadow(self._context),
            name="client_agent_external",
            socket_type=zmq.DEALER,
            address=self._scheduler_address,
            bind_or_connect="connect",
            callback=self.__on_receive_from_scheduler,
            identity=self._identity,
        )

        self._disconnect_manager: Optional[ClientDisconnectManager] = None
        self._heartbeat_manager: Optional[ClientHeartbeatManager] = None
        self._task_manager: Optional[ClientTaskManager] = None

    def __initialize(self):
        self._disconnect_manager = ClientDisconnectManager()
        self._heartbeat_manager = ClientHeartbeatManager(death_timeout_seconds=self._timeout_seconds)
        self._object_manager = ClientObjectManager(identity=self._identity)
        self._task_manager = ClientTaskManager()

        # register all managers
        self._disconnect_manager.register(
            connector_internal=self._connector_internal, connector_external=self._connector_external
        )
        self._object_manager.register(
            connector_internal=self._connector_internal, connector_external=self._connector_external
        )
        self._task_manager.register(
            connector_external=self._connector_external,
            object_manager=self._object_manager,
            future_manager=self._future_manager,
        )
        self._heartbeat_manager.register(connector_external=self._connector_external)

    def __run_loop(self):
        self._loop = asyncio.new_event_loop()
        self._task = self._loop.create_task(self.__get_loops())
        self._loop.run_until_complete(self._task)
        self._loop.close()

    def run(self):
        self.__initialize()
        self.__run_loop()

    async def __on_receive_from_client(self, message: Message):
        if isinstance(message, ClientDisconnect):
            await self._disconnect_manager.on_client_disconnect(message)
            return

        if isinstance(message, ObjectInstruction):
            await self._object_manager.on_object_instruction(message)
            return

        if isinstance(message, ObjectRequest):
            await self._object_manager.on_object_request(message)
            return

        if isinstance(message, Task):
            await self._task_manager.on_new_task(message)
            return

        if isinstance(message, TaskCancel):
            await self._task_manager.on_cancel_task(message)
            return

        if isinstance(message, GraphTask):
            await self._task_manager.on_new_graph_task(message)
            return

        if isinstance(message, GraphTaskCancel):
            await self._task_manager.on_cancel_graph_task(message)
            return

        raise TypeError(f"Unknown {message=}")

    async def __on_receive_from_scheduler(self, message: Message):
        if isinstance(message, ClientShutdownResponse):
            await self._disconnect_manager.on_client_shutdown_response(message)
            return

        if isinstance(message, ClientHeartbeatEcho):
            await self._heartbeat_manager.on_heartbeat_echo(message)
            return

        if isinstance(message, TaskResult):
            await self._task_manager.on_task_result(message)
            return

        if isinstance(message, ObjectResponse):
            self._future_manager.on_object_response(message)
            return

        raise TypeError(f"Unknown {message=}")

    async def __get_loops(self):
        await self._heartbeat_manager.send_heartbeat()

        exception = None
        try:
            await asyncio.gather(
                create_async_loop_routine(self._connector_external.routine, 0),
                create_async_loop_routine(self._connector_internal.routine, 0),
                create_async_loop_routine(self._heartbeat_manager.routine, self._heartbeat_interval_seconds),
            )
        except BaseException as e:
            exception = e
        finally:
            self._stop_event.set()  # always set the stop event before setting futures' exceptions

            await self._object_manager.clear_all_objects(clear_serializer=True)

            self._connector_external.destroy()
            self._connector_internal.destroy()

        if exception is None:
            return

        if isinstance(exception, asyncio.CancelledError):
            logging.error("ClientAgent: async. loop cancelled")
            self._future_manager.set_all_futures_with_exception(ClientCancelledException("client cancelled"))
        elif isinstance(exception, (ClientQuitException, ClientShutdownException)):
            logging.info("ClientAgent: client quitting")
            self._future_manager.set_all_futures_with_exception(exception)
        elif isinstance(exception, TimeoutError):
            logging.error(f"ClientAgent: client timeout when connecting to {self._scheduler_address.to_address()}")
            self._future_manager.set_all_futures_with_exception(exception)
        else:
            raise exception



import ctypes
import gc
import logging
import multiprocessing
import platform
import threading
import time
from typing import Any, Dict, List, Optional

import cloudpickle
import psutil

from scaler.client.serializer.mixins import Serializer
from scaler.io.config import CLEANUP_INTERVAL_SECONDS
from scaler.io.utility import concat_list_of_bytes
from scaler.protocol.python.common import ObjectContent
from scaler.protocol.python.message import Task
from scaler.utility.exceptions import DeserializeObjectError
from scaler.utility.object_utility import generate_serializer_object_id, is_object_id_serializer


class ObjectCache(threading.Thread):
    def __init__(self, garbage_collect_interval_seconds: int, trim_memory_threshold_bytes: int):
        threading.Thread.__init__(self)

        self._serializers: Dict[bytes, Serializer] = dict()

        self._garbage_collect_interval_seconds = garbage_collect_interval_seconds
        self._previous_garbage_collect_time = time.time()
        self._trim_memory_threshold_bytes = trim_memory_threshold_bytes

        self._cached_objects: Dict[bytes, Any] = {}
        self._cached_objects_alive_since: Dict[bytes, float] = dict()
        self._process = psutil.Process(multiprocessing.current_process().pid)
        self._libc = ctypes.cdll.LoadLibrary("libc.{}".format("so.6" if platform.uname()[0] != "Darwin" else "dylib"))

        self._stop_event = threading.Event()

    def run(self) -> None:
        try:
            while not self._stop_event.wait(timeout=CLEANUP_INTERVAL_SECONDS):
                self.__clean_memory()
        finally:
            self.__clear()  # gracefully destroy all cached objects

    def destroy(self) -> None:
        self._stop_event.set()

    def add_serializer(self, client: bytes, serializer: Serializer):
        self._serializers[client] = serializer

    def serialize(self, client: bytes, obj: Any) -> bytes:
        return self.get_serializer(client).serialize(obj)

    def deserialize(self, client: bytes, payload: List[bytes]) -> Any:
        return self.get_serializer(client).deserialize(concat_list_of_bytes(payload))

    def add_objects(self, object_content: ObjectContent, task: Task):
        zipped = list(zip(object_content.object_ids, object_content.object_names, object_content.object_bytes))
        serializers = filter(lambda o: is_object_id_serializer(o[0]), zipped)
        others = filter(lambda o: not is_object_id_serializer(o[0]), zipped)

        for object_id, object_name, object_bytes in serializers:
            self.add_serializer(object_id, cloudpickle.loads(concat_list_of_bytes(object_bytes)))

        for object_id, object_name, object_bytes in others:
            try:
                deserialized = self.deserialize(task.source, object_bytes)
            except Exception:  # noqa
                logging.exception(
                    f"failed to deserialize received {object_name=}, object_id={object_id.hex()}, "
                    f"length={len(object_bytes)}"
                )
                continue

            self._cached_objects[object_id] = deserialized
            self._cached_objects_alive_since[object_id] = time.time()

    def del_object(self, object_id: bytes):
        self._cached_objects_alive_since.pop(object_id, None)
        self._cached_objects.pop(object_id, None)

    def has_object(self, object_id: bytes):
        return object_id in self._cached_objects or object_id in self._serializers

    def get_object(self, object_id: bytes) -> Optional[Any]:
        if object_id not in self._cached_objects:
            raise ValueError(f"cannot get object for object_id={object_id.hex()}")

        obj = self._cached_objects[object_id]

        self._cached_objects_alive_since[object_id] = time.time()
        return obj

    def get_serializer(self, client: bytes) -> Serializer:
        serializer_id = generate_serializer_object_id(client)
        if serializer_id not in self._serializers:
            raise DeserializeObjectError(f"cannot get serializer for client={client.hex()}")

        return self._serializers[serializer_id]

    def __clean_memory(self):
        if time.time() - self._previous_garbage_collect_time < self._garbage_collect_interval_seconds:
            return

        self._previous_garbage_collect_time = time.time()

        gc.collect()

        if self._process.memory_info().rss < self._trim_memory_threshold_bytes:
            return

        self._libc.malloc_trim(0)

    def __clear(self) -> None:
        self._cached_objects.clear()
        self._cached_objects_alive_since.clear()



import logging
from typing import Dict, Optional, Set

from scaler.io.async_binder import AsyncBinder
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.common import TaskStatus
from scaler.protocol.python.message import StateTask, Task, TaskCancel, TaskResult
from scaler.protocol.python.status import TaskManagerStatus
from scaler.scheduler.graph_manager import VanillaGraphTaskManager
from scaler.scheduler.mixins import ClientManager, ObjectManager, TaskManager, WorkerManager
from scaler.utility.mixins import Looper, Reporter
from scaler.utility.queues.async_indexed_queue import AsyncIndexedQueue


class VanillaTaskManager(TaskManager, Looper, Reporter):
    def __init__(self, max_number_of_tasks_waiting: int):
        self._max_number_of_tasks_waiting = max_number_of_tasks_waiting
        self._binder: Optional[AsyncBinder] = None
        self._binder_monitor: Optional[AsyncConnector] = None

        self._client_manager: Optional[ClientManager] = None
        self._object_manager: Optional[ObjectManager] = None
        self._worker_manager: Optional[WorkerManager] = None
        self._graph_manager: Optional[VanillaGraphTaskManager] = None

        self._task_id_to_task: Dict[bytes, Task] = dict()

        self._unassigned: AsyncIndexedQueue = AsyncIndexedQueue()
        self._running: Set[bytes] = set()

        self._success_count: int = 0
        self._failed_count: int = 0
        self._canceled_count: int = 0
        self._notfound_count: int = 0
        self._no_worker_count: int = 0
        self._died_count: int = 0

    def register(
        self,
        binder: AsyncBinder,
        binder_monitor: AsyncConnector,
        client_manager: ClientManager,
        object_manager: ObjectManager,
        worker_manager: WorkerManager,
        graph_manager: VanillaGraphTaskManager,
    ):
        self._binder = binder
        self._binder_monitor = binder_monitor

        self._client_manager = client_manager
        self._object_manager = object_manager
        self._worker_manager = worker_manager
        self._graph_manager = graph_manager

    async def routine(self):
        task_id = await self._unassigned.get()

        # FIXME: As the assign_task_to_worker() call can be blocking (especially if there is no worker connected to the
        # scheduler), we might end up with the task object being in neither _running nor _unassigned.
        # In this case, the scheduler will answer any task cancellation request with a "task not found" error, which is
        # a bug.
        # https://github.com/Citi/scaler/issues/45

        if not await self._worker_manager.assign_task_to_worker(self._task_id_to_task[task_id]):
            await self._unassigned.put(task_id)
            return

        self._running.add(task_id)
        await self.__send_monitor(
            task_id,
            self._object_manager.get_object_name(self._task_id_to_task[task_id].func_object_id),
            TaskStatus.Running,
        )

    def get_status(self) -> TaskManagerStatus:
        return TaskManagerStatus.new_msg(
            unassigned=self._unassigned.qsize(),
            running=len(self._running),
            success=self._success_count,
            failed=self._failed_count,
            canceled=self._canceled_count,
            not_found=self._notfound_count,
        )

    async def on_task_new(self, client: bytes, task: Task):
        if (
            0 <= self._max_number_of_tasks_waiting <= self._unassigned.qsize()
            and not self._worker_manager.has_available_worker()
        ):
            await self._binder.send(client, TaskResult.new_msg(task.task_id, TaskStatus.NoWorker))
            return

        self._client_manager.on_task_begin(client, task.task_id)
        self._task_id_to_task[task.task_id] = task

        await self._unassigned.put(task.task_id)
        await self.__send_monitor(
            task.task_id,
            self._object_manager.get_object_name(self._task_id_to_task[task.task_id].func_object_id),
            TaskStatus.Inactive,
        )

    async def on_task_cancel(self, client: bytes, task_cancel: TaskCancel):
        if task_cancel.task_id not in self._task_id_to_task:
            logging.warning(f"cannot cancel, task not found: task_id={task_cancel.task_id.hex()}")
            await self.on_task_done(TaskResult.new_msg(task_cancel.task_id, TaskStatus.NotFound))
            return

        if task_cancel.task_id in self._unassigned:
            await self.on_task_done(TaskResult.new_msg(task_cancel.task_id, TaskStatus.Canceled))
            return

        await self._worker_manager.on_task_cancel(task_cancel)
        await self.__send_monitor(
            task_cancel.task_id,
            self._object_manager.get_object_name(self._task_id_to_task[task_cancel.task_id].func_object_id),
            TaskStatus.Canceling,
        )

    async def on_task_done(self, result: TaskResult):
        if result.status == TaskStatus.Success:
            self._success_count += 1
        elif result.status == TaskStatus.Failed:
            self._failed_count += 1
        elif result.status == TaskStatus.Canceled:
            self._canceled_count += 1
        elif result.status == TaskStatus.NotFound:
            self._notfound_count += 1
        elif result.status == TaskStatus.NoWorker:
            self._no_worker_count += 1
        elif result.status == TaskStatus.WorkerDied:
            self._died_count += 1

        if result.task_id in self._unassigned:
            self._unassigned.remove(result.task_id)
        elif result.task_id in self._running:
            self._running.remove(result.task_id)

        if result.task_id in self._task_id_to_task:
            func_object_name = self._object_manager.get_object_name(
                self._task_id_to_task.pop(result.task_id).func_object_id
            )
            client = self._client_manager.on_task_finish(result.task_id)
        else:
            func_object_name = b"<unknown func object>"
            client = None

        await self.__send_monitor(result.task_id, func_object_name, result.status, result.metadata)

        if self._graph_manager.is_graph_sub_task(result.task_id):
            await self._graph_manager.on_graph_sub_task_done(result)
            return

        if client is not None:
            await self._binder.send(client, result)

    async def on_task_reroute(self, task_id: bytes):
        assert self._client_manager.get_client_id(task_id) is not None

        self._running.remove(task_id)
        await self._unassigned.put(task_id)
        await self.__send_monitor(
            task_id,
            self._object_manager.get_object_name(self._task_id_to_task[task_id].func_object_id),
            TaskStatus.Inactive,
        )

    async def __send_monitor(
        self, task_id: bytes, function_name: bytes, status: TaskStatus, metadata: Optional[bytes] = b""
    ):
        worker = self._worker_manager.get_worker_by_task_id(task_id)
        await self._binder_monitor.send(StateTask.new_msg(task_id, function_name, status, worker, metadata))



from collections import defaultdict
from typing import Dict, Set, Tuple

from scaler.protocol.python.common import ObjectContent
from scaler.protocol.python.message import ObjectInstruction, ObjectRequest, ObjectResponse
from scaler.utility.many_to_many_dict import ManyToManyDict
from scaler.worker.agent.mixins import ObjectTracker


class VanillaObjectTracker(ObjectTracker):
    def __init__(self):
        self._object_request_to_processor_ids: ManyToManyDict[Tuple[bytes, ...], bytes] = ManyToManyDict()
        self._object_id_to_processors_ids: ManyToManyDict[bytes, bytes] = ManyToManyDict()

    def on_object_request(self, processor_id: bytes, object_request: ObjectRequest) -> None:
        self._object_request_to_processor_ids.add(object_request.object_ids, processor_id)

    def on_object_response(self, object_response: ObjectResponse) -> Set[bytes]:
        """Returns a list of processor ids that requested this object content."""

        if object_response.response_type != ObjectResponse.ObjectResponseType.Content:
            raise TypeError(f"invalid object response type received: {object_response.response_type}.")

        object_ids = object_response.object_content.object_ids

        if not self._object_request_to_processor_ids.has_left_key(object_ids):
            return set()

        processor_ids = self._object_request_to_processor_ids.remove_left_key(object_ids)

        for processor_id in processor_ids:
            for object_id in object_ids:
                self._object_id_to_processors_ids.add(object_id, processor_id)

        return processor_ids

    def on_object_instruction(self, object_instruction: ObjectInstruction) -> Dict[bytes, ObjectInstruction]:
        """
        From an object instruction received by the worker, returns the sub-object instructions that should be
        forwarded to processors.
        """

        if object_instruction.instruction_type != ObjectInstruction.ObjectInstructionType.Delete:
            raise TypeError(f"invalid object instruction type received: {object_instruction.instruction_type}.")

        per_processor_object_ids: Dict[bytes, Set[bytes]] = defaultdict(set)
        for object_id in object_instruction.object_content.object_ids:
            if not self._object_id_to_processors_ids.has_left_key(object_id):
                continue

            processor_ids = self._object_id_to_processors_ids.remove_left_key(object_id)
            for processor_id in processor_ids:
                per_processor_object_ids[processor_id].add(object_id)

        return {
            processor_id: ObjectInstruction.new_msg(
                instruction_type=ObjectInstruction.ObjectInstructionType.Delete,
                object_user=object_instruction.object_user,
                object_content=ObjectContent.new_msg(object_ids=tuple(object_ids)),
            )
            for processor_id, object_ids in per_processor_object_ids.items()
        }

    def on_processor_end(self, processor_id: bytes) -> None:
        if self._object_request_to_processor_ids.has_right_key(processor_id):
            self._object_request_to_processor_ids.remove_right_key(processor_id)

        if self._object_id_to_processors_ids.has_right_key(processor_id):
            self._object_id_to_processors_ids.remove_right_key(processor_id)



import logging
import multiprocessing
import os
import signal
from typing import List, Optional, Tuple

from scaler.utility.logging.utility import setup_logger
from scaler.utility.zmq_config import ZMQConfig
from scaler.worker.worker import Worker


class Cluster(multiprocessing.get_context("spawn").Process):  # type: ignore[misc]
    def __init__(
        self,
        address: ZMQConfig,
        worker_io_threads: int,
        worker_names: List[str],
        heartbeat_interval_seconds: int,
        task_timeout_seconds: int,
        death_timeout_seconds: int,
        garbage_collect_interval_seconds: int,
        trim_memory_threshold_bytes: int,
        hard_processor_suspend: bool,
        event_loop: str,
        logging_paths: Tuple[str, ...],
        logging_level: str,
        logging_config_file: Optional[str],
    ):
        multiprocessing.Process.__init__(self, name="WorkerMaster")

        self._address = address
        self._worker_io_threads = worker_io_threads
        self._worker_names = worker_names
        self._heartbeat_interval_seconds = heartbeat_interval_seconds
        self._task_timeout_seconds = task_timeout_seconds
        self._death_timeout_seconds = death_timeout_seconds
        self._garbage_collect_interval_seconds = garbage_collect_interval_seconds
        self._trim_memory_threshold_bytes = trim_memory_threshold_bytes
        self._hard_processor_suspend = hard_processor_suspend
        self._event_loop = event_loop

        self._logging_paths = logging_paths
        self._logging_level = logging_level
        self._logging_config_file = logging_config_file

        self._workers: List[Worker] = []

    def run(self):
        setup_logger(self._logging_paths, self._logging_config_file, self._logging_level)
        self.__register_signal()
        self.__start_workers_and_run_forever()

    def __destroy(self, *args):
        assert args is not None
        logging.info(f"{self.__get_prefix()} received signal, shutting down")
        for worker in self._workers:
            logging.info(f"{self.__get_prefix()} shutting down worker[{worker.pid}]")
            os.kill(worker.pid, signal.SIGINT)

    def __register_signal(self):
        signal.signal(signal.SIGINT, self.__destroy)
        signal.signal(signal.SIGTERM, self.__destroy)

    def __start_workers_and_run_forever(self):
        logging.info(
            f"{self.__get_prefix()} starting {len(self._worker_names)} workers, heartbeat_interval_seconds="
            f"{self._heartbeat_interval_seconds}, task_timeout_seconds={self._task_timeout_seconds}"
        )

        self._workers = [
            Worker(
                event_loop=self._event_loop,
                name=name,
                address=self._address,
                io_threads=self._worker_io_threads,
                heartbeat_interval_seconds=self._heartbeat_interval_seconds,
                garbage_collect_interval_seconds=self._garbage_collect_interval_seconds,
                trim_memory_threshold_bytes=self._trim_memory_threshold_bytes,
                task_timeout_seconds=self._task_timeout_seconds,
                death_timeout_seconds=self._death_timeout_seconds,
                hard_processor_suspend=self._hard_processor_suspend,
                logging_paths=self._logging_paths,
                logging_level=self._logging_level,
            )
            for name in self._worker_names
        ]

        for worker in self._workers:
            worker.start()

        for i, worker in enumerate(self._workers):
            logging.info(f"Worker[{worker.ident}] started")

        for worker in self._workers:
            worker.join()

        logging.info(f"{self.__get_prefix()} shutdown")

    def __get_prefix(self):
        return f"{self.__class__.__name__}:"



import time
import unittest

from tests.utility import get_available_tcp_port, logging_test_name

from scaler import Client, SchedulerClusterCombo
from scaler.utility.logging.utility import setup_logger


def dummy(n: int):
    time.sleep(n)
    return n * n


def busy_dummy(n: int):
    start_time = time.time()
    while time.time() - start_time < n:
        pass

    return n * n


class TestProfiling(unittest.TestCase):
    def setUp(self):
        setup_logger()
        logging_test_name(self)
        self.address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        self.cluster = SchedulerClusterCombo(
            address=self.address, n_workers=2, per_worker_queue_size=2, event_loop="builtin"
        )
        self.client = Client(address=self.address, profiling=True)
        self.client_off = Client(address=self.address, profiling=False)

    def tearDown(self) -> None:
        self.client.disconnect()
        self.client_off.disconnect()
        self.cluster.shutdown()

    def test_future_completed(self) -> None:
        fut = self.client.submit(dummy, 1)
        fut.result()

        task_time = fut.profiling_info().duration_s
        assert task_time > 0

    def test_future_incomplete(self) -> None:
        fut = self.client.submit(dummy, 1, profiling=True)

        # Raises error because future is not done
        with self.assertRaises(ValueError):
            _ = fut.profiling_info().duration_s

    def test_cpu_time_busy(self) -> None:
        fut = self.client.submit(busy_dummy, 1, profiling=True)
        fut.result()

        cpu_time = fut.profiling_info().cpu_time_s
        assert cpu_time > 0

    def test_cpu_time_sleep(self) -> None:
        fut = self.client.submit(dummy, 1, profiling=True)
        fut.result()

        cpu_time = fut.profiling_info().cpu_time_s
        assert cpu_time < 1



import collections
import graphlib
import itertools
from typing import Generic, Hashable, Iterable, List, Mapping, Optional, Tuple, TypeVar

from bidict import bidict

try:
    import graphblas as gb
    import numpy as np  # noqa
except ImportError:
    raise ImportError("Please use 'pip install python-graphblas' to have graph blas support")

GraphKeyType = TypeVar("GraphKeyType", bound=Hashable)


class TopologicalSorter(Generic[GraphKeyType]):
    """
    Implements graphlib's TopologicalSorter, but the graph handling is backed by GraphBLAS
    Reference: https://github.com/python/cpython/blob/4a3ea1fdd890e5e2ec26540dc3c958a52fba6556/Lib/graphlib.py
    """

    def __init__(self, graph: Optional[Mapping[GraphKeyType, Iterable[GraphKeyType]]] = None):
        # the layout of the matrix is (in-vertex, out-vertex)
        self._matrix = gb.Matrix(gb.dtypes.BOOL)
        self._key_to_id: bidict[GraphKeyType, int] = bidict()

        self._graph_matrix_mask: Optional[np.ndarray] = None
        self._visited_vertices_mask: Optional[np.ndarray] = None
        self._ready_nodes: Optional[List[GraphKeyType]] = None

        self._n_done = 0
        self._n_visited = 0

        if graph is not None:
            self.merge_graph(graph)

    def add(self, node: GraphKeyType, *predecessors: GraphKeyType) -> None:
        self.merge_graph({node: predecessors})

    def merge_graph(self, graph: Mapping[GraphKeyType, Iterable[GraphKeyType]]) -> None:
        if self._ready_nodes is not None:
            raise ValueError("nodes cannot be added after a call to prepare()")

        # cache old dim to compare later when resizing matrix
        old_dim = len(self._key_to_id)

        # maintain iterable copies for iterable predecessors
        graph_iterable_copy = {}

        # update key to id mappings
        for node, predecessors in graph.items():
            if node not in self._key_to_id:
                self._key_to_id[node] = len(self._key_to_id)

            # copy iterator if predecessors is an iterable
            if isinstance(predecessors, collections.abc.Iterable):
                predecessors, graph_iterable_copy[node] = itertools.tee(predecessors)

            for pred in predecessors:
                if pred not in self._key_to_id:
                    self._key_to_id[pred] = len(self._key_to_id)

        # resize at once as it is faster
        if old_dim != len(self._key_to_id):
            self._matrix.resize(len(self._key_to_id), len(self._key_to_id))

        # update matrix
        for node, predecessors in graph.items():
            if node in graph_iterable_copy:
                predecessors = graph_iterable_copy[node]

            for pred in predecessors:
                self._matrix[self._key_to_id[node], self._key_to_id[pred]] = True

    def prepare(self) -> None:
        if self._ready_nodes is not None:
            raise ValueError("cannot prepare() more than once")

        self._graph_matrix_mask = np.ones(len(self._key_to_id), bool)
        self._visited_vertices_mask = np.zeros(len(self._key_to_id), bool)

        self._ready_nodes = self._get_zero_degree_keys()
        for node in self._ready_nodes:
            self._visited_vertices_mask[self._key_to_id[node]] = True
        self._n_visited += len(self._ready_nodes)

        if self._has_cycle():
            raise graphlib.CycleError("cycle detected")

    def get_ready(self) -> Tuple[GraphKeyType, ...]:
        if self._ready_nodes is None:
            raise ValueError("prepare() must be called first")

        result = tuple(self._ready_nodes)
        self._ready_nodes.clear()
        return result

    def is_active(self) -> bool:
        if self._ready_nodes is None:
            raise ValueError("prepare() must be called first")
        return self._n_done < self._n_visited or bool(self._ready_nodes)

    def __bool__(self) -> bool:
        return self.is_active()

    def done(self, *nodes: GraphKeyType) -> None:
        if self._ready_nodes is None:
            raise ValueError("prepare() must be called first")

        for node in nodes:
            if node not in self._key_to_id:
                raise ValueError(f"node {node!r} was not added using add()")

            _id = self._key_to_id[node]

            if not self._visited_vertices_mask[_id]:
                raise ValueError(f"node {node!r} is not ready")

            if not self._graph_matrix_mask[_id]:
                raise ValueError(f"node {node!r} is already done")

            self._graph_matrix_mask[_id] = False
        self._n_done += len(nodes)

        new_ready_nodes = self._get_zero_degree_keys()
        for node in new_ready_nodes:
            self._visited_vertices_mask[self._key_to_id[node]] = True
        self._ready_nodes.extend(new_ready_nodes)
        self._n_visited += len(new_ready_nodes)

    def static_order(self) -> Iterable[GraphKeyType]:
        self.prepare()
        while self.is_active():
            node_group = self.get_ready()
            yield from node_group
            self.done(*node_group)

    def _has_cycle(self) -> bool:
        """
        Detect cycle using trace(A^n) != 0.
        https://arxiv.org/pdf/1610.01200.pdf

        :return: True if cycle is found, otherwise False
        """
        matrix_n = gb.Vector.from_dense(np.ones(len(self._key_to_id), bool), missing_value=False).diag()
        for _ in range(len(self._key_to_id)):
            # use LOR_PAIR to compute matrix multiplication over boolean matrices
            matrix_n << gb.semiring.lor_pair(matrix_n @ self._matrix)
            # check diagonal for any truthy values
            if matrix_n.diag().reduce(gb.monoid.lor):
                return True
        return False

    def _get_zero_degree_keys(self) -> List[GraphKeyType]:
        ids = self._get_mask_diff(self._visited_vertices_mask, self._get_zero_degree_mask(self._get_masked_matrix()))
        return [self._key_to_id.inverse[_id] for _id in ids]

    def _get_masked_matrix(self) -> gb.Matrix:
        # convert vector mask to matrix diagonal and then perform matrix multiplication to mask matrix
        # https://github.com/DrTimothyAldenDavis/GraphBLAS/issues/48#issuecomment-858596341
        return gb.semiring.lor_pair(
            self._matrix @ gb.Vector.from_dense(self._graph_matrix_mask, missing_value=False).diag()
        )

    @classmethod
    def _get_zero_degree_mask(cls, masked_matrix: gb.Matrix) -> np.ndarray:
        degrees = masked_matrix.reduce_rowwise(gb.monoid.lor)
        indices, _ = degrees.to_coo(indices=True, values=False, sort=False)
        return np.logical_not(np.in1d(np.arange(masked_matrix.nrows), indices))  # type: ignore[attr-defined]

    @staticmethod
    def _get_mask_diff(old_mask: np.ndarray, new_mask: np.ndarray) -> List[int]:
        return np.argwhere(old_mask != new_mask).ravel().tolist()



class TaskNotFoundError(Exception):
    pass


class WorkerDiedError(Exception):
    pass


class NoWorkerError(Exception):
    pass


class DisconnectedError(Exception):
    pass


class ProcessorDiedError(Exception):
    pass


class DeserializeObjectError(Exception):
    pass


class MissingObjects(Exception):
    pass


class ClientCancelledException(Exception):
    pass


class ClientShutdownException(Exception):
    pass


class ClientQuitException(Exception):
    pass
